#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
import time
from binascii import hexlify

sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                                    "..")))

from pasaffe_lib.readdb import PassSafeFile  # noqa: E402


class TestReadDB(unittest.TestCase):
    def setUp(self):
        self.passfile = PassSafeFile()

    def test_folder_list_to_field(self):

        folder_list = [[[], ""],
                       [["foldera"], "foldera"],
                       [["folder.a"], "folder\.a"],  # noqa: W605
                       [["foldera."], "foldera\."],  # noqa: W605
                       [[".foldera"], "\.foldera"],  # noqa: W605
                       [["foldera.", "folderb."],
                         "foldera\..folderb\."],  # noqa: W605
                       [["foldera", "folderb"], "foldera.folderb"],
                       [["folder.a", "folderb"],
                         "folder\.a.folderb"],  # noqa: W605
                       [["foldera", "folder.b"], "foldera.folder\.b"],
                       [["folder.a", "folder.b"], "folder\.a.folder\.b"],
                       [["folder.a", "folder.b", "folder.c"],
                        "folder\.a.folder\.b.folder\.c"],
                       [["", "foldera", "folderb"], ".foldera.folderb"],
                       [["", "", "folderb"], "..folderb"],
                       [["foldera", "", "folderb"], "foldera..folderb"],
                       [["foldera", "folderb", ""], "foldera.folderb."],
                       ]

        for (folder, field) in folder_list:
            self.assertEqual(self.passfile._folder_list_to_field(folder),
                             field)

    def test_field_to_folder_list(self):

        folder_list = [["", []],
                       ["foldera", ["foldera"]],
                       ["folder\.a", ["folder.a"]],  # noqa: W605
                       ["foldera\.", ["foldera."]],  # noqa: W605
                       ["\.foldera", [".foldera"]],  # noqa: W605
                       ["foldera\..folderb\.",
                        ["foldera.", "folderb."]],  # noqa: W605
                       ["foldera.folderb", ["foldera", "folderb"]],
                       ["folder\.a.folderb", ["folder.a", "folderb"]],
                       ["foldera.folder\.b", ["foldera", "folder.b"]],
                       ["folder\.a.folder\.b", ["folder.a", "folder.b"]],
                       ["folder\.a.folder\.b.folder\.c",
                       ["folder.a", "folder.b", "folder.c"]],
                       [".foldera.folderb", ["", "foldera", "folderb"]],
                       ["..folderb", ["", "", "folderb"]],
                       ["foldera..folderb", ["foldera", "", "folderb"]],
                       ["foldera.folderb.", ["foldera", "folderb", ""]],
                       ]

        for (field, folder) in folder_list:
            self.assertEqual(self.passfile._field_to_folder_list(field),
                             folder)

    def test_get_database_version_string(self):

        self.passfile.new_db("test")

        expected = '%s.%s' % (
            hexlify(self.passfile.db_version[1:2]).decode('utf-8'),
            hexlify(self.passfile.db_version[0:1]).decode('utf-8'))

        self.assertEqual(self.passfile.get_database_version_string(), expected)

        self.passfile.header[0] = b'\x0B\x03'

        self.assertEqual(self.passfile.get_database_version_string(), "03.0b")

    def test_new_entry(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        self.assertEqual(len(uuid_hex), 32)
        self.assertTrue(uuid_hex in self.passfile.records)

        for field in [1, 3, 4, 5, 6, 7, 8, 12, 13]:
            self.assertTrue(field in self.passfile.records[uuid_hex])

    def test_delete_entry(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        self.assertTrue(uuid_hex in self.passfile.records)

        self.passfile.delete_entry(uuid_hex)

        self.assertTrue(uuid_hex not in self.passfile.records)

    def test_update_modification_time(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        old_time = self.passfile.records[uuid_hex][12]
        time.sleep(1.1)
        self.passfile.update_modification_time(uuid_hex)
        new_time = self.passfile.records[uuid_hex][12]

        self.assertTrue(old_time != new_time)

    def test_update_password_time(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        old_time = self.passfile.records[uuid_hex][8]
        time.sleep(1.1)
        self.passfile.update_password_time(uuid_hex)
        new_time = self.passfile.records[uuid_hex][8]

        self.assertTrue(old_time != new_time)

    def test_get_title(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        test_title = "Test title"
        self.passfile.records[uuid_hex][3] = test_title

        title = self.passfile.get_title(uuid_hex)
        self.assertTrue(title == test_title)

    def test_get_username(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        test_username = "awesomeuser"
        self.passfile.records[uuid_hex][4] = test_username

        username = self.passfile.get_username(uuid_hex)
        self.assertTrue(username == test_username)

    def test_get_notes(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        test_note = "This is a note."
        self.passfile.records[uuid_hex][5] = test_note

        note = self.passfile.get_notes(uuid_hex)
        self.assertTrue(note == test_note)

    def test_get_password(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        test_password = "s3cr3tp4ssw0rd"
        self.passfile.records[uuid_hex][6] = test_password

        password = self.passfile.get_password(uuid_hex)
        self.assertTrue(password == test_password)

    def test_get_url(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        test_url = "http://www.example.com"
        self.passfile.records[uuid_hex][13] = test_url

        url = self.passfile.get_url(uuid_hex)
        self.assertTrue(url == test_url)

    def test_update_folder_list(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        self.assertTrue(2 not in self.passfile.records[uuid_hex])

        folder = ['folderA', 'folderB', 'folderC']
        folder_field = 'folderA.folderB.folderC'

        self.passfile.update_folder_list(uuid_hex, folder)

        self.assertTrue(2 in self.passfile.records[uuid_hex])
        self.assertTrue(self.passfile.records[uuid_hex][2] == folder_field)

        self.passfile.update_folder_list(uuid_hex, [])
        self.assertTrue(2 not in self.passfile.records[uuid_hex])

    def test_get_folder_list(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        self.assertTrue(2 not in self.passfile.records[uuid_hex])
        self.assertTrue(self.passfile.get_folder_list(uuid_hex) == [])

        folder = ['folderA', 'folderB', 'folderC']
        folder_field = 'folderA.folderB.folderC'

        self.passfile.update_folder_list(uuid_hex, folder)

        self.assertTrue(2 in self.passfile.records[uuid_hex])
        self.assertTrue(self.passfile.records[uuid_hex][2] == folder_field)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex) == folder)

        self.assertTrue(self.passfile.get_folder_list('nonexistent') == [])

    def test_get_empty_folders(self):

        folder_fields = ['folderA',
                         'folderA.folderB',
                         'folderA.folderB.folderC']

        folder_list = [['folderA'],
                       ['folderA', 'folderB'],
                       ['folderA', 'folderB', 'folderC']]

        self.assertTrue(self.passfile.get_empty_folders() == [])
        self.passfile.empty_folders = folder_fields

        self.assertTrue(self.passfile.get_empty_folders() == folder_list)

    def test_add_empty_folder(self):

        folder_fields = ['folderA',
                         'folderA.folderB',
                         'folderA.folderB.folderC']

        folder = ['folderA', 'folderB', 'folderC']

        # Make sure it's empty
        self.assertTrue(self.passfile.empty_folders == [])

        # Add a folder, and make sure it created the children
        self.passfile.add_empty_folder(folder)
        self.assertTrue(self.passfile.empty_folders == folder_fields)

        # Make sure empty parameters work
        self.passfile.add_empty_folder(None)
        self.assertTrue(self.passfile.empty_folders == folder_fields)
        self.passfile.add_empty_folder([])
        self.assertTrue(self.passfile.empty_folders == folder_fields)

        # Try adding it again, to make sure we don't have duplicates
        self.passfile.add_empty_folder(folder)
        self.assertTrue(self.passfile.empty_folders == folder_fields)

        # Make sure adding an empty folder that isn't actually empty
        # works
        uuid_hex = self.passfile.new_entry()
        entry_folder = ['otherA']
        self.passfile.update_folder_list(uuid_hex, entry_folder)
        self.passfile.add_empty_folder(entry_folder)
        self.assertTrue(self.passfile.empty_folders == folder_fields)

        # Adding an empty folder that only has an empty subfolder should
        # only add the subfolder
        uuid_hex = self.passfile.new_entry()
        entry_folder = ['thirdA', 'thirdB']
        self.passfile.update_folder_list(uuid_hex, entry_folder)
        self.passfile.add_empty_folder(entry_folder)

        folder_fields.append('thirdA')
        self.assertTrue(self.passfile.empty_folders == folder_fields)

        # Do the same, but for 3 levels
        uuid_hex = self.passfile.new_entry()
        entry_folder = ['fourthA', 'fourthB', 'fourthC']
        self.passfile.update_folder_list(uuid_hex, entry_folder)
        self.passfile.add_empty_folder(entry_folder)

        folder_fields.append('fourthA')
        folder_fields.append('fourthA.fourthB')
        self.assertTrue(self.passfile.empty_folders == folder_fields)

    def test_remove_empty_folder(self):

        folder_fields = ['folderA',
                         'folderA.folderB',
                         'folderA.folderB.folderC']

        # pass by value
        self.passfile.empty_folders = folder_fields[:]

        # Try and remove a bogus folder
        self.passfile.remove_empty_folder(['bogus'])
        self.assertTrue(self.passfile.empty_folders == folder_fields)

        # Make sure empty parameters work
        self.passfile.remove_empty_folder(None)
        self.assertTrue(self.passfile.empty_folders == folder_fields)
        self.passfile.remove_empty_folder([])
        self.assertTrue(self.passfile.empty_folders == folder_fields)

        # Now, remove an empty folder
        self.passfile.remove_empty_folder(['folderA', 'folderB'])
        folder_fields.remove('folderA.folderB')
        self.assertTrue(self.passfile.empty_folders == folder_fields)

    def test_get_all_folders(self):
        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        # First make sure there's no folders at all
        self.assertTrue(self.passfile.get_all_folders() == [])

        # Now handle an entry, but no empty folders
        folder = ['folderA', 'folderB']
        self.passfile.update_folder_list(uuid_hex, folder)
        self.assertTrue(self.passfile.get_all_folders() == [folder])

        # Now handle empty folders, but no entry
        self.passfile.update_folder_list(uuid_hex, [])
        self.assertTrue(self.passfile.get_all_folders() == [])

        folder_fields = ['folderA',
                         'folderA.folderB',
                         'folderA.folderB.folderC']
        self.passfile.empty_folders = folder_fields

        folder_list = [['folderA'],
                       ['folderA', 'folderB'],
                       ['folderA', 'folderB', 'folderC']]

        self.assertTrue(self.passfile.get_all_folders() == folder_list)

        # Now handle multiple entries, and empty folders
        self.passfile.update_folder_list(uuid_hex, folder)

        folderB = ['OtherFolderA', 'OtherFolderB']

        uuid_hex_B = self.passfile.new_entry()
        self.passfile.new_entry()
        self.passfile.update_folder_list(uuid_hex_B, folderB)

        all_folders = folder_list + [folderB]
        self.assertTrue(self.passfile.get_all_folders() == all_folders)

    def test_rename_folder_list(self):

        self.passfile.new_db("test")

        # Create a few entries
        folderA = ['firstA', 'firstB']
        folderB = ['firstA', 'firstB', 'firstC']
        folderC = ['secondA', 'secondB']

        uuid_hex_A = self.passfile.new_entry()
        uuid_hex_B = self.passfile.new_entry()
        uuid_hex_C = self.passfile.new_entry()
        uuid_hex_D = self.passfile.new_entry()

        self.passfile.update_folder_list(uuid_hex_A, folderA)
        self.passfile.update_folder_list(uuid_hex_B, folderB)
        self.passfile.update_folder_list(uuid_hex_C, folderC)

        self.assertTrue(self.passfile.get_folder_list(uuid_hex_A) == folderA)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_B) == folderB)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_C) == folderC)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_D) == [])

        # Ok, let's rename one
        new_folderA = ['firstA', 'renamedB']
        new_folderB = ['firstA', 'renamedB', 'firstC']
        self.passfile.rename_folder_list(folderA, new_folderA)

        self.assertTrue(
            self.passfile.get_folder_list(uuid_hex_A) == new_folderA)
        self.assertTrue(
            self.passfile.get_folder_list(uuid_hex_B) == new_folderB)
        self.assertTrue(
            self.passfile.get_folder_list(uuid_hex_C) == folderC)
        self.assertTrue(
            self.passfile.get_folder_list(uuid_hex_D) == [])

        # Now add some empty_folders
        folder_fields = ['firstA',
                         'secondA',
                         'thirdA',
                         'thirdA.thirdB']

        # pass by value
        self.passfile.empty_folders = folder_fields[:]

        # OK, rename entries and empty_folders
        another_new_folderA = ['renamedA', 'renamedB']
        another_new_folderB = ['renamedA', 'renamedB', 'firstC']
        new_empty_folders = [['secondA'],
                             ['thirdA'],
                             ['thirdA', 'thirdB'],
                             ['renamedA']]

        self.passfile.rename_folder_list(['firstA'], ['renamedA'])

        self.assertTrue(
            self.passfile.get_folder_list(uuid_hex_A) == another_new_folderA)
        self.assertTrue(
            self.passfile.get_folder_list(uuid_hex_B) == another_new_folderB)
        self.assertTrue(
            self.passfile.get_folder_list(uuid_hex_C) == folderC)
        self.assertTrue(
            self.passfile.get_folder_list(uuid_hex_D) == [])

        self.assertTrue(
            self.passfile.get_empty_folders() == new_empty_folders)

    def test_delete_folder(self):

        self.passfile.new_db("test")

        # Create a few entries
        folderA = ['firstA', 'firstB']
        folderB = ['firstA', 'firstB', 'firstC']
        folderC = ['secondA', 'secondB']

        uuid_hex_A = self.passfile.new_entry()
        uuid_hex_B = self.passfile.new_entry()
        uuid_hex_C = self.passfile.new_entry()
        uuid_hex_D = self.passfile.new_entry()

        self.passfile.update_folder_list(uuid_hex_A, folderA)
        self.passfile.update_folder_list(uuid_hex_B, folderB)
        self.passfile.update_folder_list(uuid_hex_C, folderC)

        self.assertTrue(self.passfile.get_folder_list(uuid_hex_A) == folderA)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_B) == folderB)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_C) == folderC)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_D) == [])

        self.assertTrue(self.passfile.get_empty_folders() == [])

        # Try and delete some invalid things
        self.passfile.delete_folder(None)
        self.passfile.delete_folder([])
        self.passfile.delete_folder(['banana'])

        self.assertTrue(self.passfile.get_folder_list(uuid_hex_A) == folderA)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_B) == folderB)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_C) == folderC)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_D) == [])

        self.assertTrue(self.passfile.get_empty_folders() == [])

        # Now, delete a third level.
        self.passfile.delete_folder(folderB)

        self.assertTrue(self.passfile.get_folder_list(uuid_hex_A) == folderA)
        self.assertTrue(uuid_hex_B not in self.passfile.records)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_C) == folderC)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_D) == [])
        self.assertTrue(uuid_hex_D in self.passfile.records)
        self.assertTrue(self.passfile.get_empty_folders() == [])

        # Add some empty_folders and delete a toplevel
        emptyA = [['firstA'],
                  ['firstA', 'firstB', 'emptyC']]
        emptyB = emptyA + [['some'],
                           ['some', 'random']]
        for folder in emptyB:
            self.passfile.add_empty_folder(folder)

        self.assertTrue(self.passfile.get_empty_folders() == emptyB)
        self.passfile.delete_folder(['some'])
        self.assertTrue(self.passfile.get_empty_folders() == emptyA)

        self.assertTrue(self.passfile.get_folder_list(uuid_hex_A) == folderA)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_C) == folderC)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_D) == [])
        self.assertTrue(uuid_hex_D in self.passfile.records)

        # Now delete a top level
        self.passfile.delete_folder(['firstA'])
        self.assertTrue(self.passfile.get_empty_folders() == [])
        self.assertTrue(uuid_hex_A not in self.passfile.records)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_C) == folderC)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_D) == [])
        self.assertTrue(uuid_hex_D in self.passfile.records)

        # Delete the remaining second level. We should gain an empty folder.
        self.passfile.delete_folder(['secondA', 'secondB'])
        self.assertTrue(uuid_hex_C not in self.passfile.records)
        self.assertTrue(self.passfile.get_folder_list(uuid_hex_D) == [])
        self.assertTrue(uuid_hex_D in self.passfile.records)
        self.assertTrue(self.passfile.get_empty_folders() == [['secondA']])

    def test_get_tree_status(self):

        self.passfile.new_db("test")
        self.assertEqual(self.passfile.get_tree_status(), None)

        self.passfile.header[6] = "Pasaffe test"
        self.assertEqual(self.passfile.get_tree_status(), None)

        self.passfile.header[3] = "101010"
        self.assertEqual(self.passfile.get_tree_status(), "101010")

        self.passfile.header[6] = "AnotherApp v1"
        self.assertEqual(self.passfile.get_tree_status(), None)

    def test_set_tree_status(self):

        self.passfile.new_db("test")
        self.passfile.header[6] = "Pasaffe test"
        self.assertEqual(self.passfile.get_tree_status(), None)

        self.passfile.set_tree_status("101010")
        self.assertEqual(self.passfile.get_tree_status(), "101010")
        self.assertTrue(3 in self.passfile.header)

        self.passfile.set_tree_status(None)
        self.assertEqual(self.passfile.get_tree_status(), None)
        self.assertFalse(3 in self.passfile.header)

    def test_fixups(self):

        self.passfile.new_db("test")
        uuid_hex = self.passfile.new_entry()

        # New entry should have empty strings
        self.assertEqual(self.passfile.records[uuid_hex][3], '')
        self.assertEqual(self.passfile.records[uuid_hex][4], '')
        self.assertEqual(self.passfile.records[uuid_hex][5], '')
        self.assertEqual(self.passfile.records[uuid_hex][6], '')

        # Delete title, username and password
        del self.passfile.records[uuid_hex][3]
        del self.passfile.records[uuid_hex][4]
        del self.passfile.records[uuid_hex][6]

        # add a comment with CRLF terminators
        crlf = ("First line\r\n" +
                "Second line\r\n" +
                "Third line")

        lf = ("First line\n" +
              "Second line\n" +
              "Third line")

        self.passfile.records[uuid_hex][5] = crlf

        self.assertTrue(3 not in self.passfile.records[uuid_hex])
        self.assertTrue(4 not in self.passfile.records[uuid_hex])
        self.assertTrue(6 not in self.passfile.records[uuid_hex])
        self.assertEqual(self.passfile.records[uuid_hex][5], crlf)

        # Now do the postread fixup
        self.passfile._postread_fixup()

        self.assertEqual(self.passfile.records[uuid_hex][3], '')
        self.assertEqual(self.passfile.records[uuid_hex][4], '')
        self.assertEqual(self.passfile.records[uuid_hex][5], lf)
        self.assertEqual(self.passfile.records[uuid_hex][6], '')

        # OK, do the presave fixup
        self.assertEqual(self.passfile._presave_fixup(uuid_hex, 5), crlf)

    def _create_find_db(self):

        entries = [{3: "carte de crédit",
                    4: "username1",
                    5: "This is a note",
                    6: "password1",
                    13: "http://www.example.com"},
                   {2: "folder1",
                    3: "carte de credit",
                    4: "username1",
                    5: "anothernote",
                    6: "password1"},
                   {2: "level1group.level2group.level3group",
                    3: "level3entry",
                    4: "usernamelevel3",
                    6: "passwordlevel3",
                    13: "http://note.com"}]

        self.passfile.new_db('pasaffe')

        for entry in entries:
            uuid = self.passfile.new_entry()
            for key in entry.keys():
                self.passfile.records[uuid][key] = entry[key]

    def test_update_find_results(self):

        self._create_find_db()

        self.passfile.update_find_results("")
        self.assertEqual(self.passfile.find_results, [])
        self.assertEqual(self.passfile.find_results_index, None)
        self.assertEqual(self.passfile.find_value, "")

        # Test field 3
        self.passfile.update_find_results("level3entry")
        self.assertEqual(len(self.passfile.find_results), 1)
        self.assertEqual(self.passfile.find_results_index, None)
        self.assertEqual(self.passfile.find_value, "level3entry")

        # Test field 5
        self.passfile.update_find_results("another")
        self.assertEqual(len(self.passfile.find_results), 1)
        self.assertEqual(self.passfile.find_results_index, None)
        self.assertEqual(self.passfile.find_value, "another")

        # Test field 13
        self.passfile.update_find_results("example.com")
        self.assertEqual(len(self.passfile.find_results), 1)
        self.assertEqual(self.passfile.find_results_index, None)
        self.assertEqual(self.passfile.find_value, "example.com")

        # Test matching entries with accents
        self.passfile.update_find_results("credit")
        self.assertEqual(len(self.passfile.find_results), 2)
        self.assertEqual(self.passfile.find_results_index, None)
        self.assertEqual(self.passfile.find_value, "credit")

        # Test specifying search term with accent
        self.passfile.update_find_results("crédit")
        self.assertEqual(len(self.passfile.find_results), 2)
        self.assertEqual(self.passfile.find_results_index, None)
        self.assertEqual(self.passfile.find_value, "crédit")

        # Test force option
        self.passfile.find_results = []
        self.passfile.update_find_results("crédit")
        self.assertEqual(len(self.passfile.find_results), 0)
        self.passfile.update_find_results("crédit", True)
        self.assertEqual(len(self.passfile.find_results), 2)

        # Test clearing out values
        self.passfile.update_find_results("")
        self.assertEqual(self.passfile.find_results, [])
        self.assertEqual(self.passfile.find_results_index, None)
        self.assertEqual(self.passfile.find_value, "")

    def test_get_next_find_result(self):

        self._create_find_db()

        self.assertEqual(self.passfile.get_next_find_result(), None)
        self.assertEqual(self.passfile.get_next_find_result(True), None)

        self.passfile.update_find_results("note")
        self.assertEqual(len(self.passfile.find_results), 3)
        self.assertEqual(self.passfile.find_results_index, None)
        self.assertEqual(self.passfile.find_value, "note")

        self.assertEqual(self.passfile.find_results_index, None)

        uuid1 = self.passfile.get_next_find_result()
        self.assertEqual(self.passfile.find_results_index, 0)
        self.assertNotEqual(uuid1, None)

        uuid2 = self.passfile.get_next_find_result()
        self.assertEqual(self.passfile.find_results_index, 1)
        self.assertNotEqual(uuid2, None)
        self.assertNotEqual(uuid1, uuid2)

        uuid3 = self.passfile.get_next_find_result()
        self.assertEqual(self.passfile.find_results_index, 2)
        self.assertNotEqual(uuid3, None)
        self.assertNotEqual(uuid1, uuid3)
        self.assertNotEqual(uuid2, uuid3)

        # This should wrap around
        uuid4 = self.passfile.get_next_find_result()
        self.assertEqual(self.passfile.find_results_index, 0)
        self.assertNotEqual(uuid4, None)
        self.assertEqual(uuid1, uuid4)

        # And now try backwards
        uuid5 = self.passfile.get_next_find_result(True)
        self.assertEqual(self.passfile.find_results_index, 2)
        self.assertNotEqual(uuid5, None)
        self.assertEqual(uuid3, uuid5)


if __name__ == '__main__':
    unittest.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import optparse

import gettext
t = gettext.translation('pasaffe', fallback=True)
_ = t.gettext

import gi  # noqa: E402
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, Gtk  # noqa: E402

from pasaffe import PasaffeWindow  # noqa: E402

from pasaffe_lib import set_up_logging, get_version  # noqa: E402
from pasaffe_lib.helpers import get_database_path  # noqa: E402


def parse_options():
    """Support for command line options"""
    parser = optparse.OptionParser(usage="Usage: %prog [filename] [options]",
                                   version="%%prog %s" % get_version())
    parser.add_option(
        "-v", "--verbose", action="count", dest="verbose",
        help=_("Show debug messages (-vv debugs pasaffe_lib also)"))
    parser.add_option(
        "-f", "--file", dest="filename",
        help=_("use database FILE"), metavar="FILE")
    parser.add_option(
        "-s", "--set-default", dest="set_default",
        help=_("set database as default"), action='store_true')
    (options, args) = parser.parse_args()
    if args and options.filename is None:
        options.filename = args[0]
    set_up_logging(options)
    return options


def main():
    'constructor for your class instances'
    options = parse_options()

    filename = get_database_path()

    settings = Gio.Settings.new("net.launchpad.pasaffe")

    # On first launch, set the standard location
    if settings.get_string('database-path') == "":
        settings.set_string('database-path', filename)

    # Override path that was saved with path from command line
    if options.set_default and options.filename is not None:
        settings.set_string('database-path', options.filename)

    # Run the application.
    window = PasaffeWindow.PasaffeWindow(database=options.filename)
    window.show()  # pylint: disable=E1101
    Gtk.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import struct
import hashlib
import hmac
import random
import re
import os
import time
import tempfile
import shutil
if os.name == "posix":
    import pwd
elif os.name == "nt":
    import getpass
# import pwd
from binascii import hexlify  # noqa: E402
from unidecode import unidecode  # noqa: E402
from pasaffe_lib.helpers import PathEntry  # noqa: E402

from . import pytwofishcbc  # noqa: E402
import logging  # noqa: E402
logger = logging.getLogger('pasaffe_lib')
from . pasaffeconfig import get_version  # noqa: E402


class PassSafeFile:

    def __init__(self, filename=None, password=None, req_cipher='Twofish',
                 fixup=True):
        '''Reads a Password Safe v3 file'''

        self.keys = {}
        self.header = {}
        self.records = {}
        self.cipher = None
        self.cipher_block_size = 0
        self.hmac = None
        self.dbfile = None
        self.empty_folders = []

        self.find_results = []
        self.find_results_index = None
        self.find_value = ""

        # These fields need converting between strings and bytes
        self.header_text = [0x03, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x11]
        self.record_text = [0x02, 0x03, 0x04, 0x05, 0x06, 0x0d]

        # Use version 0x030B, since we support saving empty folders
        self.db_version = b'\x0B\x03'

        if req_cipher == 'Twofish':
            self.cipher = pytwofishcbc.TwofishCBC()
            self.cipher_block_size = self.cipher.get_block_size()
        else:
            raise ValueError("Sorry, we don't support %s yet." % req_cipher)

        if filename is not None:
            self.readfile(filename, password, fixup=fixup)

    def readfile(self, filename, password, fixup=True):
        '''Parses database file'''
        logger.debug('Opening database: %s' % filename)
        try:
            self.dbfile = open(filename, 'rb')
        except Exception:
            raise RuntimeError("Could not open %s. Aborting." % filename)

        tag = self.dbfile.read(4)
        if tag != b"PWS3":
            raise RuntimeError("File %s is not a password safe database."
                               " Aborting." % filename)

        self._readkeys(password)
        self._readheader()
        self._readrecords()
        self._validatehmac()
        self.dbfile.close()
        self.dbfile = None

        if fixup:
            self._postread_fixup()

        # Now that we've read the file, but before we get rid of the
        # password, generate new keys for our next save
        self.new_keys(password)

        # Don't need the password anymore, clear it out
        password = ''

    def new_db(self, password):
        '''Creates a new database in memory'''
        self.keys['ITER'] = 2048

        self.new_keys(password)

        self.header[0] = self.db_version
        self.header[1] = os.urandom(16)  # uuid

    def check_password(self, password):
        '''Checks if password is valid'''
        stretched_key = self._keystretch(password,
                                         self.keys['SALT'],
                                         self.keys['ITER'])
        # Don't need the password anymore, clear it out
        password = ''
        if hashlib.sha256(stretched_key).digest() != self.keys['HP']:
            return False
        else:
            return True

    def new_keys(self, password):
        '''Generates new keys'''
        self.keys['SALT'] = os.urandom(32)

        stretched_key = self._keystretch(password, self.keys['SALT'],
                                         self.keys['ITER'])
        # Don't need the password anymore, clear it out
        password = ''
        self.keys['HP'] = hashlib.sha256(stretched_key).digest()
        self.cipher.set_key(stretched_key)
        # Don't need the stretched key anymore, clear it out
        stretched_key = ''

        b1_rand = os.urandom(16)
        b2_rand = os.urandom(16)
        b3_rand = os.urandom(16)
        b4_rand = os.urandom(16)
        self.keys['K'] = b1_rand + b2_rand
        self.keys['L'] = b3_rand + b4_rand
        self.keys['B1'] = self.cipher.encrypt(b1_rand)
        self.keys['B2'] = self.cipher.encrypt(b2_rand)
        self.keys['B3'] = self.cipher.encrypt(b3_rand)
        self.keys['B4'] = self.cipher.encrypt(b4_rand)
        self.keys['IV'] = os.urandom(16)

    def writefile(self, filename, backup=False):
        '''Writes database file'''

        # Set username
        if os.name == "posix":
            self.header[7] = pwd.getpwuid(os.getuid())[0]
        elif os.name == "nt":
            self.header[7] = getpass.getuser()
        # Remove the old deprecated username field if it exists
        if 5 in self.header:
            del self.header[5]
        # Set hostname
        self.header[8] = os.uname()[1]
        # Set timestamp
        self.header[4] = struct.pack("<I", int(time.time()))
        self.header[6] = "Pasaffe v%s" % get_version()

        # Update the database version if it's older than what we support.
        # Do major version first
        if self.header[0][1] < self.db_version[1]:
            self.header[0] = self.db_version
        # Now check minor version if major version is same
        elif self.header[0][1] == self.db_version[1]:
            if self.header[0][0] < self.db_version[0]:
                self.header[0] = self.db_version

        # Create backup if requested
        if backup is True and os.path.exists(filename):
            shutil.copy(filename, filename + ".bak")

        basedir = os.path.dirname(filename)

        try:
            self.dbfile = tempfile.NamedTemporaryFile(dir=basedir,
                                                      delete=False)
        except Exception:
            raise RuntimeError("Could not create %s. Aborting." % filename)

        tempname = self.dbfile.name

        self.dbfile.write(b"PWS3")
        self._writekeys()
        self._writeheader()
        self._writerecords()
        self._writeeofblock()
        self._writehmac()
        self.dbfile.close()
        self.dbfile = None

        # TODO: add sanity check
        # At this point, intention was to reopen the temp file and see
        # if we can parse it before copying it over as a sanity check,
        # but we don't have the password anymore

        # Copy it over the real database
        shutil.copy(tempname, filename)
        os.unlink(tempname)

    def get_database_version_string(self):
        '''Returns a string of the current database version'''
        return '%s.%s' % (hexlify(self.header[0][1:2]).decode('utf-8'),
                          hexlify(self.header[0][0:1]).decode('utf-8'))

    def get_saved_name(self):
        '''Returns the username of the last save'''
        return self.header.get(7)

    def get_saved_host(self):
        '''Returns the hostname of the last save'''
        return self.header.get(8)

    def get_saved_application(self):
        '''Returns the application of the last save'''
        return self.header.get(6)

    def get_saved_date_string(self, localtime=True):
        '''Returns a string of the date of the last save'''
        if 4 in self.header:
            unpacked_time = struct.unpack("<I", self.header[4])[0]
            if localtime:
                converted_time = time.localtime(unpacked_time)
            else:
                converted_time = time.gmtime(unpacked_time)
            return time.strftime("%a, %d %b %Y %H:%M:%S", converted_time)
        else:
            return None

    def get_tree_status(self):
        '''Returns the tree display status'''
        # Tree display status is implementation specific
        if self.header.get(6, "").startswith("Pasaffe") and 3 in self.header:
            return self.header.get(3)
        else:
            return None

    def set_tree_status(self, status):
        '''Sets the tree display status'''
        if status is None:
            if 3 in self.header:
                del self.header[3]
        else:
            self.header[3] = status

    def get_folder_list(self, uuid):
        '''Returns a list of folders an entry belongs to'''
        if uuid not in self.records:
            return []

        if 2 not in self.records[uuid]:
            return []

        return (self._field_to_folder_list(self.records[uuid][2]))

    def update_folder_list(self, uuid, folder):
        '''Updates an entry folder list'''
        if uuid not in self.records:
            return

        self.records[uuid][2] = self._folder_list_to_field(folder)

        # If the record is empty, just delete it
        if self.records[uuid][2] == "":
            del self.records[uuid][2]

    def get_empty_folders(self):
        '''Returns the empty folders list'''
        folders = []

        for folder in self.empty_folders:
            folders.append(self._field_to_folder_list(folder))

        logger.debug("returning %s" % folders)
        return folders

    def get_all_folders(self):
        '''Returns a list of all the folders'''

        # First, get the empty folders
        folders = self.get_empty_folders()

        # Now, do the records
        for uuid in self.records:
            if 2 not in self.records[uuid]:
                continue

            folder = self._field_to_folder_list(self.records[uuid][2])
            if folder not in folders:
                folders.append(folder)

        return folders

    def add_empty_folder(self, folder):
        '''Adds a folder to the empty folders list'''

        if folder is None or folder == []:
            return

        for part in range(len(folder)):
            field = self._folder_list_to_field(folder[:part + 1])
            logger.debug("searching for %s" % field)
            if field not in self.empty_folders:
                # Make sure it's actually empty
                found = False
                for uuid in list(self.records.keys()):
                    if 2 not in self.records[uuid]:
                        continue
                    if self.records[uuid][2] == field:
                        logger.debug("folder %s isn't empty" % field)
                        found = True
                        break

                if found is False:
                    logger.debug("adding %s" % field)
                    self.empty_folders.append(field)

    def remove_empty_folder(self, folder):
        '''Removes a folder from the empty folders list'''
        field = self._folder_list_to_field(folder)
        if field in self.empty_folders:
            logger.debug("removing %s" % field)
            self.empty_folders.remove(field)

    def rename_folder_list(self, old_list, new_list):
        '''Renamed a folder name in all entries'''
        old_field = self._folder_list_to_field(old_list)
        new_field = self._folder_list_to_field(new_list)

        # Do the records first
        for uuid in self.records:
            if 2 not in self.records[uuid]:
                continue

            if self.records[uuid][2] == old_field:
                self.records[uuid][2] = new_field
            elif self.records[uuid][2].startswith(old_field + '.'):
                updated_field = self.records[uuid][2].replace(
                    old_field, new_field, 1)
                self.records[uuid][2] = updated_field
            else:
                continue

            self.update_modification_time(uuid)

        # Now do the empty folders
        for empty_folder in self.empty_folders[:]:
            if empty_folder == old_field:
                logger.debug("renaming %s to %s" % (empty_folder, new_field))
                self.empty_folders.remove(empty_folder)
                self.empty_folders.append(new_field)
            elif empty_folder.startswith(old_field + '.'):
                updated_field = empty_folder.replace(old_field, new_field, 1)
                logger.debug("renaming %s to %s" % (empty_folder,
                                                    updated_field))
                self.empty_folders.remove(empty_folder)
                self.empty_folders.append(updated_field)

    def delete_entry(self, uuid):
        '''Deletes an entry'''
        del self.records[uuid]

    def delete_folder(self, folder):
        '''Deletes a folder and all contents'''

        if folder is None or folder == []:
            return

        field = self._folder_list_to_field(folder)

        # Do the records first
        for uuid in list(self.records.keys()):
            if 2 not in self.records[uuid]:
                continue
            if self.records[uuid][2] == field:
                self.delete_entry(uuid)
            elif self.records[uuid][2].startswith(field + '.'):
                self.delete_entry(uuid)

        # Now do the empty folders
        for empty_folder in self.empty_folders[:]:
            if empty_folder == field:
                logger.debug("removing folder '%s'" % empty_folder)
                self.empty_folders.remove(empty_folder)
            elif empty_folder.startswith(field + '.'):
                logger.debug("removing folder '%s'" % empty_folder)
                self.empty_folders.remove(empty_folder)

        # If the parent folder has no contents,
        # add it to empty folders list
        parent = folder[:-1]
        if parent == []:
            return
        parent_field = self._folder_list_to_field(parent)

        for uuid in self.records:
            if 2 not in self.records[uuid]:
                continue
            if self.records[uuid][2] == parent_field:
                return

        self.add_empty_folder(parent)

    def update_modification_time(self, uuid):
        '''Updates the modification time of an entry'''
        timestamp = struct.pack("<I", int(time.time()))
        self.records[uuid][12] = timestamp

    def get_title(self, uuid):
        '''Returns the entry title'''
        return self.records[uuid].get(3)

    def get_username(self, uuid):
        '''Returns the entry username'''
        return self.records[uuid].get(4)

    def get_notes(self, uuid):
        '''Returns the entry notes'''
        return self.records[uuid].get(5)

    def get_password(self, uuid):
        '''Returns the entry password'''
        return self.records[uuid].get(6)

    def get_url(self, uuid):
        '''Returns the entry URL'''
        return self.records[uuid].get(13)

    def get_modification_time(self, uuid, localtime=True):
        '''Returns a string of the entry modification time'''
        return self.get_time(uuid, 12, localtime)

    def update_password_time(self, uuid):
        '''Updates the password time of an entry'''
        timestamp = struct.pack("<I", int(time.time()))
        self.records[uuid][8] = timestamp

    def get_password_time(self, uuid, localtime=True):
        '''Returns a string of the password modification time'''
        return self.get_time(uuid, 8, localtime)

    def get_creation_time(self, uuid, localtime=True):
        '''Returns a string of the creation time'''
        return self.get_time(uuid, 7, localtime)

    def get_time(self, uuid, entry, localtime=True):
        '''Returns a string of time'''
        if entry in self.records[uuid]:
            unpacked_time = struct.unpack("<I", self.records[uuid][entry])[0]
            if localtime:
                converted_time = time.localtime(unpacked_time)
            else:
                converted_time = time.gmtime(unpacked_time)
            return time.strftime("%a, %d %b %Y %H:%M:%S", converted_time)
        else:
            return None

    def _folder_list_to_field(self, folder_list):
        '''Converts a folder list to a folder field'''
        field = ""

        if folder_list is None:
            return field

        have_folders = False
        for folder in folder_list:
            if have_folders is True:
                field += "."
            field += folder.replace(".", "\\.")
            have_folders = True
        return field

    def _field_to_folder_list(self, field):
        '''Converts a folder field to a folder list'''

        # We need to split into folders using the "." character, but not
        # if it is escaped with a \
        folders = []

        if field == "":
            return folders

        index = 0
        location = 0

        while index < len(field):

            if field[index] == ".":
                folders.append("")
                location += 1
                index += 1
                continue

            location = field.find(".", location + 1)

            if location == -1:
                break

            if field[location - 1] == "\\":
                continue

            folders.append(field[index:location].replace("\\", ''))
            index = location + 1

        folders.append(field[index:len(field)].replace('\\', ''))
        return folders

    def new_entry(self):
        '''Creates a new entry'''
        uuid = os.urandom(16)
        uuid_hex = hexlify(uuid).decode('utf-8')
        timestamp = struct.pack("<I", int(time.time()))
        new_entry = {1: uuid, 3: '', 4: '', 5: '', 6: '',
                     7: timestamp, 8: timestamp, 12: timestamp, 13: ''}
        self.records[uuid_hex] = new_entry

        return uuid_hex

    def _keystretch(self, password, salt, iters):
        '''Takes a password, and stretches it using iters iterations'''
        password = hashlib.sha256(password.encode('utf-8') + salt).digest()
        for i in range(iters):
            password = hashlib.sha256(password).digest()
        return password

    def _readkeys(self, password):
        self.keys['SALT'] = self.dbfile.read(32)
        self.keys['ITER'] = struct.unpack("<i", self.dbfile.read(4))[0]
        # Sanity check so we don't gobble up massive amounts of ram
        if self.keys['ITER'] > 100000:
            raise RuntimeError("Too many iterations: %s. Aborting." %
                               self.keys['ITER'])
        logger.debug("Number of iters is %d" % self.keys['ITER'])
        self.keys['HP'] = self.dbfile.read(32)
        # logger.debug("hp is %s" % self.keys['HP'])
        self.keys['B1'] = self.dbfile.read(16)
        self.keys['B2'] = self.dbfile.read(16)
        self.keys['B3'] = self.dbfile.read(16)
        self.keys['B4'] = self.dbfile.read(16)
        self.keys['IV'] = self.dbfile.read(16)
        self.cipher.initCBC(self.keys['IV'])
        stretched_key = self._keystretch(password, self.keys['SALT'],
                                         self.keys['ITER'])
        # Don't need the password anymore, clear it out
        password = ''
        # logger.debug("stretched pass is %s" % hexlify(stretched_key))
        if hashlib.sha256(stretched_key).digest() != self.keys['HP']:
            raise ValueError("Password supplied doesn't match database."
                             " Aborting.")

        self.cipher.set_key(stretched_key)
        # Don't need the stretched key anymore, clear it out
        stretched_key = b''
        self.keys['K'] = self.cipher.decrypt(self.keys['B1']) + \
            self.cipher.decrypt(self.keys['B2'])
        self.keys['L'] = self.cipher.decrypt(self.keys['B3']) + \
            self.cipher.decrypt(self.keys['B4'])
        self.hmac = hmac.new(self.keys['L'], digestmod=hashlib.sha256)
        # logger.debug("K is %s and L is %s" % (hexlify(self.keys['K']),
        #                                       hexlify(self.keys['L']))

    def _writekeys(self):
        self.dbfile.write(self.keys['SALT'])
        self.dbfile.write(struct.pack("i", self.keys['ITER']))
        self.dbfile.write(self.keys['HP'])
        self.dbfile.write(self.keys['B1'])
        self.dbfile.write(self.keys['B2'])
        self.dbfile.write(self.keys['B3'])
        self.dbfile.write(self.keys['B4'])
        self.dbfile.write(self.keys['IV'])
        self.cipher.initCBC(self.keys['IV'])
        self.hmac = hmac.new(self.keys['L'], digestmod=hashlib.sha256)

    def _readheader(self):
        self.cipher.set_key(self.keys['K'])

        while(1):
            status, field_type, field_data = self._readfield()
            if status is False:
                raise RuntimeError("Malformed file, "
                                   "was expecting more data in header")

            # Convert from bytes to strings
            if field_type in self.header_text:
                field_data = field_data.decode('utf-8')

            if field_type == 0xff:
                logger.debug("Found end field")
                break
            elif field_type == 0x11:
                # Empty group fields can appear more than once
                # Store them in their own variable
                self.empty_folders.append(field_data)
                logger.debug("found empty folder: %s" % field_data)
            else:
                self.header[field_type] = field_data
                logger.debug("Found field 0x%.2x" % field_type)

    def _writeheader(self):
        self.cipher.set_key(self.keys['K'])

        # v3.30 of the spec says the version type needs to be the
        # first field in the header. Handle it first.
        logger.debug("Writing Version Type field")
        self._writefield(0x00, self.header[0x00])

        for entry in list(self.header.keys()):
            # Skip Version Type, we've already handled it
            if entry == 0x00:
                continue

            # Convert from strings to bytes
            if entry in self.header_text:
                value = self.header[entry].encode('utf-8')
            else:
                value = self.header[entry]

            logger.debug("Writing %.2x" % entry)
            self._writefield(entry, value)

        # Now handle empty folders
        logger.debug("Writing empty folders")
        for folder in self.empty_folders:
            logger.debug("writing empty folder: %s" % folder)
            self._writefield(0x11, folder.encode('utf-8'))

        self._writefieldend()

    def _readrecords(self):
        self.cipher.set_key(self.keys['K'])

        record = {}

        while(1):
            status, field_type, field_data = self._readfield()
            if status is False:
                break
            if field_type == 0xff:
                logger.debug("Found end field")
                uuid = hexlify(record[1]).decode('utf-8')
                self.records[uuid] = record
                record = {}
            else:

                # Convert from bytes to strings
                if field_type in self.record_text:
                    field_data = field_data.decode('utf-8')

                record[field_type] = field_data
                logger.debug("Found field 0x%.2x" % field_type)

    def _writerecords(self):
        self.cipher.set_key(self.keys['K'])

        for uuid in self.records:
            for field in list(self.records[uuid].keys()):

                # Fix up some of the fields
                fixed_value = self._presave_fixup(uuid, field)

                # Convert from strings to bytes
                if field in self.record_text:
                    value = fixed_value.encode('utf-8')
                else:
                    value = fixed_value

                self._writefield(field, value)
            self._writefieldend()

    def _readfield(self):
        field_data = b''
        status, first_block = self._readblock()
        if status is False:
            return False, 0xFF, b''
        field_length = struct.unpack("<I", first_block[0:4])[0]
        field_type = struct.unpack("B", first_block[4:5])[0]

        logger.debug("field length is %d" % field_length)
        logger.debug("field_type is 0x%.2x" % field_type)

        # Do we need multiple blocks?
        if field_length <= self.cipher_block_size - 5:
            logger.debug("single block")
            field_data = first_block[5:5 + field_length]
        else:
            field_data = first_block[5:self.cipher_block_size]
            field_length -= self.cipher_block_size - 5
            while field_length > 0:
                logger.debug("extra block")
                status, data = self._readblock()
                if status is False:
                    raise RuntimeError("Malformed file, "
                                       "was expecting more data")
                field_data += data[0:field_length]
                field_length -= self.cipher_block_size

        logger.debug("actual data length is %d" % len(field_data))

        self.hmac.update(field_data)

        return True, field_type, field_data

    def _writefield(self, field_type, field_data):
        self.hmac.update(field_data)
        field_length = len(field_data)
        field_free_space = self.cipher_block_size - 5
        index = 0
        block = b''
        block += struct.pack("I", field_length)
        block += struct.pack("B", field_type)

        logger.debug("Writing field type %.2x, length %d" %
                     (field_type, field_length))

        while field_length >= 0:
            if field_length < field_free_space:
                logger.debug("smaller than block")
                block += field_data[index:index + field_length]
                for x in range(field_free_space - field_length):
                    block += struct.pack("B", random.randint(0, 254))
                self._writeblock(block)
                field_length = -1
                block = b''
            else:
                logger.debug("bigger than block")
                block += field_data[index:index + field_free_space]
                self._writeblock(block)
                field_length -= field_free_space
                if field_length == 0:
                    field_length = -1
                index += field_free_space
                field_free_space = self.cipher_block_size
                block = b''

    def _writefieldend(self):
        block = struct.pack("I", 0)
        block += struct.pack("B", 0xff)

        logger.debug("Writing field end")

        for x in range(self.cipher_block_size - 5):
            block += struct.pack("B", random.randint(0, 254))
        self._writeblock(block)

    def _readblock(self):
        block = self.dbfile.read(self.cipher_block_size)
        if block == b'PWS3-EOFPWS3-EOF':
            return False, block
        return True, self.cipher.decryptCBC(block)

    def _writeblock(self, block):
        logger.debug("writing block, length is %d" % len(block))
        self.dbfile.write(self.cipher.encryptCBC(block))

    def _writeeofblock(self):
        self.dbfile.write(b'PWS3-EOFPWS3-EOF')

    def _validatehmac(self):
        hmac = self.dbfile.read(32)
        if hmac != self.hmac.digest():
            raise RuntimeError("Malformed file, HMAC didn't match!")
        else:
            logger.debug("HMAC Matched")
            self.hmac = None

    def _writehmac(self):
        self.dbfile.write(self.hmac.digest())
        self.hmac = None

    def _postread_fixup(self):
        '''Performs some cleanup after reading certain databases'''

        for uuid in self.records:
            # Some apps don't create username fields, so let's default
            # to what the real Password Safe does and fix it up to be a
            # field with an empty string
            if 4 not in self.records[uuid]:
                self.records[uuid][4] = ''

            # Do basically the same with password fields
            if 6 not in self.records[uuid]:
                self.records[uuid][6] = ''

            # And the same with title fields
            if 3 not in self.records[uuid]:
                self.records[uuid][3] = ''

            # Most apps use CRLF line terminators. Convert them to LF
            # when opening in Pasaffe, we'll convert them back to CRLF
            # before saving
            if 5 in self.records[uuid]:
                self.records[uuid][5] = self.records[uuid][5].replace(
                    "\r\n", "\n")

    def _presave_fixup(self, uuid, field):
        '''Performs some cleanup before saving certain databases'''

        if field == 5:
            # Most apps use CRLF line terminators. Convert LF to CRLF
            # before saving
            return self.records[uuid][5].replace("\n", "\r\n")
        else:
            return self.records[uuid][field]

    def update_find_results(self, find, force=False):

        if find == "":
            self.find_results = []
            self.find_results_index = None
            self.find_value = ""
            return

        if find == self.find_value and force is False:
            return

        find_ascii = unidecode(find)
        if find_ascii != find:
            pat = re.compile(find + "|" + find_ascii, re.IGNORECASE)
        else:
            pat = re.compile(find, re.IGNORECASE)

        record_list = (3, 5, 13)
        results = []

        for uuid in self.records:
            found = False
            for record_type in record_list:
                if record_type in self.records[uuid]:
                    if pat.search(
                            self.records[uuid].get(record_type)):
                        found = True
                        break
                    if pat.search(unidecode(
                            self.records[uuid].get(record_type))):
                        found = True
                        break

            if found is True:
                entry = PathEntry(
                    self.records[uuid][3],
                    uuid,
                    self.get_folder_list(uuid))
                results.append(entry)

        self.find_results = sorted(results)
        self.find_results_index = None
        self.find_value = find

    def get_next_find_result(self, backwards=False):

        if len(self.find_results) == 0:
            return None

        if self.find_results_index is None:
            self.find_results_index = 0
        elif backwards is False:
            self.find_results_index += 1
            if self.find_results_index == len(self.find_results):
                self.find_results_index = 0
        else:
            if self.find_results_index == 0:
                self.find_results_index = len(self.find_results) - 1
            else:
                self.find_results_index -= 1

        return self.find_results[self.find_results_index].uuid



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2015 C de-Avillez <hggdh2@ubuntu.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MErcHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
import subprocess
import tempfile
import shutil

myPath = os.path.dirname(__file__)
sys.path.insert(0, os.path.realpath(os.path.join(myPath,
                                                 "..",
                                                 "../bin",
                                                 "../lib")))

# from pasaffe_lib.readdb import PassSafeFile


def cliCmd():
    return "%s/../bin/pasaffe-cli" % myPath


def subproc(command, test):
    rc = 0
    result = None
    try:
        result = subprocess.check_output(command).splitlines()
    except subprocess.CalledProcessError as err:
        print("%s failed with rc=%s" % (test, err.returncode))
        print("%s" % err.output)
        rc = err.returncode
        result = None
    return result, rc


def gen_pswd(size=16):
    command = ("%s" % cliCmd(), "--genpswd", "--pswdlen=%s" % size)
    db_pswd, rc = subproc(command, "gen_pswd")
    if rc == 0:
        db_pswd = db_pswd[0].decode('utf-8').strip()
    return db_pswd, rc


def createDB(filename, password):
    command = (cliCmd(), "--createdb", "--masterpassword=%s" % password,
               "--file=%s" % filename)
    result, rc = subproc(command, "createDB")
    return result, rc


def creatEntry(suffix):
    entry = "Entry%s" % suffix
    group = "Group%s" % suffix
    userId = "userId%s" % suffix
    passwd, rc = gen_pswd()
    password = passwd
    url = "https://url%s.com" % suffix
    notes = "Note%s line 1\nNote%s line 2" % (suffix, suffix)
    return [entry, group, userId, password, url, notes]


def addEntry(db, pswd, entry=None, group=None, userid=None, password=None,
             url=None, notes=None):
    command = []
    command.append(cliCmd())
    # command.append("--debug")
    command.append("--add")
    command.append("--file=%s" % db)
    command.append("--masterpassword=%s" % pswd)
    command.append("--entry=%s" % entry)
    if group is not None:
        command.append("--group=%s" % group)
    if userid is not None:
        command.append("--user=%s" % userid)
    if password is not None:
        command.append("--pswd=%s" % password)
    if url is not None:
        command.append("--url=%s" % url)
    if notes is not None:
        command.append("--notes=%s" % notes)
    print("command=%s" % command)
    result, rc = subproc(command, "addEntry")
    return result, rc


def replEntry(db, pswd, entry=None, newentry=None, group=None, userid=None,
              password=None, url=None, notes=None):
    command = []
    command.append(cliCmd())
    # command.append("--debug")
    command.append("--repl")
    command.append("--file=%s" % db)
    command.append("--masterpassword=%s" % pswd)
    command.append("--entry=%s" % entry)
    if newentry is not None:
        command.append("--newentry=%s" % newentry)
    if group is not None:
        command.append("--group=%s" % group)
    if userid is not None:
        command.append("--user=%s" % userid)
    if password is not None:
        command.append("--pswd=%s" % password)
    if url is not None:
        command.append("--url=%s" % url)
    if notes is not None:
        command.append("--notes=%s" % notes)
    print("command=%s" % command)
    result, rc = subproc(command, "replEntry")
    return result, rc


def listEntry(db, pswd, entry="Dummy", fuzzy=False, lstUserId=False,
              lstPswd=False, lstGroup=False, lstURL=False, lstNotes=False,
              lstAll=False):
    command = []
    command.append(cliCmd())
    # command.append("--debug")
    command.append("--file=%s" % db)
    command.append("--masterpassword=%s" % pswd)
    command.append("--entry=%s" % entry)
    if lstAll:
        command.append("--listall")
    else:
        command.append("--list")
        if fuzzy:
            command.append("--fuzzy")
        if lstUserId:
            command.append("--listuser")
        if lstGroup:
            command.append("--listgroup")
        if lstPswd:
            command.append("--listpswd")
        if lstURL:
            command.append("--listurl")
        if lstNotes:
            command.append("--listnotes")
    print("command=%s" % command)
    result, rc = subproc(command, "listEntry")
    return result, rc


class TestPasaffeCLI(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.db_name = os.path.join(self.tempdir, 'test_pasaffe-cli.psafe3')

    def tearDown(self):
        if os.path.exists(self.tempdir):
            shutil.rmtree(self.tempdir)

    def test_t01GenPassword(self):
        Len = 8
        while Len < 32:
            password, rc = gen_pswd(size=Len)
            self.assertEqual(rc, 0, msg="call to genpswd failed")
            self.assertNotEqual(len(password), 0,
                                msg="Failed to generate a password")
            self.assertEqual(len(password), Len,
                             msg="Length of generated password does not" +
                                 " match requested length")
            Len += 1

    def test_t02createDB(self):
        db_pswd, rc = gen_pswd()
        self.assertEqual(rc, 0, msg="call to genpswd failed")
        result, rc = createDB(self.db_name, db_pswd)
        self.assertEqual(rc, 0, msg="DB creation failed")
        self.assertTrue(os.path.isfile(self.db_name))

    def test_t03addEntry(self):
        db_pswd, rc = gen_pswd()
        self.assertEqual(rc, 0, msg="call to genpswd failed")
        result, rc = createDB(self.db_name, db_pswd)
        self.assertEqual(rc, 0, msg="DB creation failed")
        values = creatEntry(1)
        result, rc = addEntry(self.db_name, db_pswd, entry=values[0])
        self.assertNotEqual(rc, 0)
        values = creatEntry(2)
        result, rc = addEntry(self.db_name, db_pswd, entry=values[0],
                              group=values[1])
        self.assertNotEqual(rc, 0)
        values = creatEntry(3)
        result, rc = addEntry(self.db_name, db_pswd, entry=values[0],
                              group=values[1],
                              userid=values[2])
        self.assertEqual(rc, 0)
        values = creatEntry(4)
        result, rc = addEntry(self.db_name, db_pswd, entry=values[0],
                              group=values[1],
                              userid=values[2],
                              password=values[3])
        self.assertEqual(rc, 0)
        values = creatEntry(5)
        result, rc = addEntry(self.db_name, db_pswd, entry=values[0],
                              group=values[1],
                              userid=values[2],
                              password=values[3],
                              url=values[4])
        self.assertEqual(rc, 0)
        values = creatEntry(6)
        result, rc = addEntry(self.db_name, db_pswd, entry=values[0],
                              group=values[1],
                              userid=values[2],
                              password=values[3],
                              url=values[4],
                              notes=values[5])
        self.assertEqual(rc, 0)

    def test_t04replaceEntry(self):
        db_pswd, rc = gen_pswd()
        self.assertEqual(rc, 0, msg="call to genpswd failed")
        result, rc = createDB(self.db_name, db_pswd)
        self.assertEqual(rc, 0, msg="DB creation failed")
        values = creatEntry(1)
        result, rc = addEntry(self.db_name, db_pswd, entry=values[0],
                              group=values[1],
                              userid=values[2])
        self.assertEqual(rc, 0, msg="adding entry 1 failed")
        result, rc = replEntry(self.db_name, db_pswd, entry=values[0],
                               group=values[1],
                               newentry="NewEntry1")
        self.assertEqual(rc, 0, msg="replacing entry string failed")
        result, rc = replEntry(self.db_name, db_pswd, entry="NewEntry1",
                               group="NewGroup1")
        self.assertEqual(rc, 0, msg="replacing group string failed")
        result, rc = replEntry(self.db_name, db_pswd, entry="NewEntry1",
                               group="Group1",
                               userid="NewUser1")
        self.assertEqual(rc, 0, msg="replacing user string failed")
        result, rc = replEntry(self.db_name, db_pswd, entry="NewEntry1",
                               group="Group1",
                               userid="User1",
                               password="Password1")
        self.assertEqual(rc, 0, msg="replacing password string failed")
        result, rc = replEntry(self.db_name, db_pswd, entry="NewEntry1",
                               group="Group1",
                               userid="User1",
                               password="Password1",
                               url="http://127.0.0.1")
        self.assertEqual(rc, 0, msg="replacing URL string failed")
        result, rc = replEntry(self.db_name, db_pswd, entry="NewEntry1",
                               group="Group1",
                               userid="User1",
                               password="Password1",
                               url="http://127.0.0.1",
                               notes="This is a note for newEntry1")
        self.assertEqual(rc, 0, msg="replacing notes string failed")

    def test_t05listEntry(self):
        db_pswd, rc = gen_pswd()
        self.assertEqual(rc, 0, msg="call to genpswd failed")
        result, rc = createDB(self.db_name, db_pswd)
        self.assertEqual(rc, 0, msg="DB creation failed")
        result, rc = addEntry(self.db_name, db_pswd, entry="Entry1",
                              group="Group1", userid="User1",
                              password="Password1", url="http://127.0.0.1",
                              notes="This is a note for Entry1")
        self.assertEqual(rc, 0, msg="adding entry 1 failed")
        result, rc = addEntry(self.db_name, db_pswd, entry="Entry2",
                              group="Group2", userid="User2",
                              password="Password2", url="http://127.0.0.2",
                              notes="This is a note for Entry2")
        self.assertEqual(rc, 0, msg="adding entry 2 failed")
        result, rc = addEntry(self.db_name, db_pswd, entry="Entry3",
                              group="Group3", userid="User3",
                              password="Password3", url="http://127.0.0.3",
                              notes="This is a note for Entry3")
        self.assertEqual(rc, 0, msg="adding entry 3 failed")
        result, rc = listEntry(self.db_name, db_pswd, entry="Entry1")
        self.assertEqual(rc, 0, "list Entry1 1/n failed")
        print("Result:%s" % result)
        result, rc = listEntry(self.db_name, db_pswd, entry="Entry2", )
        self.assertEqual(rc, 0, "list Entry1 2/n failed")
        result, rc = listEntry(self.db_name, db_pswd, entry="Entry3", )
        self.assertEqual(rc, 0, "list Entry1 2/n failed")

    def test_t06removeEntry(self):
        pass


if __name__ == '__main__':
    unittest.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2012 Francesco Marella <fra.marella@gmx.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import struct
import os
import time
from binascii import hexlify
from xml.etree import cElementTree as ET
import logging
logger = logging.getLogger('pasaffe_lib')


class FigaroXML:
    def __init__(self, filename=None):
        """ Reads a FPM2 file"""

        self.records = {}
        self.skipped = []
        self.index = 0
        self.cipher = None

        if filename is not None:
            self.readfile(filename)

    def readfile(self, filename):
        """ Parses database file"""
        try:
            element = ET.parse(filename)
        except Exception:
            raise RuntimeError("Could not open %s. Aborting." % filename)

        if element.getroot().tag != 'FPM':
            raise RuntimeError("Not a valid FPM2 XML file")

        for pwitem in element.findall('./PasswordList/PasswordItem'):
            uuid = os.urandom(16)
            uuid_hex = hexlify(uuid).decode('utf-8')
            timestamp = struct.pack("<I", int(time.time()))
            new_entry = {1: uuid, 3: '', 4: '', 6: '',
                         7: timestamp, 8: timestamp, 12: timestamp}

            for x in list(pwitem):
                if x.tag == 'title':
                    new_entry[3] = (x.text or 'Untitled item')
                elif x.tag == 'user':
                    new_entry[4] = (x.text or '')
                elif x.tag == 'password':
                    new_entry[6] = (x.text or '')
                elif x.tag == 'url':
                    new_entry[13] = (x.text or '')
                elif x.tag == 'notes':
                    new_entry[5] = (x.text or '')
                elif x.tag == 'category':
                    category = x.text
                    if category is not None:
                        new_entry[2] = category.replace(".", "\\.")
                else:
                    if x.tag not in self.skipped:
                        self.skipped.append(x.tag)

            self.records[uuid_hex] = new_entry



# twofish.py - pure Python implementation of the Twofish algorithm.
# Bjorn Edstrom <be@bjrn.se> 13 december 2007.
#
# Copyrights
# ==========
#
# This code is a derived from an implementation by Dr Brian Gladman
# (gladman@seven77.demon.co.uk) which is subject to the following license.
# This Python implementation is not subject to any other license.
#
# /* This is an independent implementation of the encryption algorithm:   */
# /*                                                                      */
# /*         Twofish by Bruce Schneier and colleagues                     */
# /*                                                                      */
# /* which is a candidate algorithm in the Advanced Encryption Standard   */
# /* programme of the US National Institute of Standards and Technology.  */
# /*                                                                      */
# /* Copyright in this implementation is held by Dr B R Gladman but I     */
# /* hereby give permission for its free direct or derivative use subject */
# /* to acknowledgment of its origin and compliance with any conditions   */
# /* that the originators of t he algorithm place on its exploitation.    */
# /*                                                                      */
# /* My thanks to Doug Whiting and Niels Ferguson for comments that led   */
# /* to improvements in this implementation.                              */
# /*                                                                      */
# /* Dr Brian Gladman (gladman@seven77.demon.co.uk) 14th January 1999     */
#
# The above copyright notice must not be removed.
#
# Information
# ===========
#
# Anyone thinking of using this code should reconsider. It's slow.
# Try python-mcrypt instead. In case a faster library is not installed
# on the target system, this code can be used as a portable fallback.

try:
    import psyco
    psyco.full()
except ImportError:
    pass

import struct
import sys

block_size = 16
key_size = 32


class Twofish:

    def __init__(self, key=None):
        """Twofish."""

        if key:
            self.set_key(key)

    def set_key(self, key):
        """Init."""

        key_len = len(key)
        if key_len not in [16, 24, 32]:
            # XXX: add padding?
            raise KeyError("key must be 16, 24 or 32 bytes")
        if key_len % 4:
            # XXX: add padding?
            raise KeyError("key not a multiple of 4")
        if key_len > 32:
            # XXX: prune?
            raise KeyError("key_len > 32")

        self.context = TWI()

        key_word32 = [0] * 32
        i = 0
        while key:
            key_word32[i] = struct.unpack("<L", key[0:4])[0]
            key = key[4:]
            i += 1

        set_key(self.context, key_word32, key_len)

    def decrypt(self, block):
        """Decrypt blocks."""

        if len(block) % 16:
            raise ValueError("block size must be a multiple of 16")

        plaintext = b''

        while block:
            a, b, c, d = struct.unpack("<4L", block[:16])
            temp = [a, b, c, d]
            decrypt(self.context, temp)
            plaintext += struct.pack("<4L", *temp)
            block = block[16:]

        return plaintext

    def encrypt(self, block):
        """Encrypt blocks."""

        if len(block) % 16:
            raise ValueError("block size must be a multiple of 16")

        ciphertext = b''

        while block:
            a, b, c, d = struct.unpack("<4L", block[0:16])
            temp = [a, b, c, d]
            encrypt(self.context, temp)
            ciphertext += struct.pack("<4L", *temp)
            block = block[16:]

        return ciphertext

    def get_name(self):
        """Return the name of the cipher."""

        return "Twofish"

    def get_block_size(self):
        """Get cipher block size in bytes."""

        return 16

    def get_key_size(self):
        """Get cipher key size in bytes."""

        return 32


#
# Private.
#

WORD_BIGENDIAN = 0
if sys.byteorder == 'big':
    WORD_BIGENDIAN = 1


def rotr32(x, n):
    return (x >> n) | ((x << (32 - n)) & 0xFFFFFFFF)


def rotl32(x, n):
    return ((x << n) & 0xFFFFFFFF) | (x >> (32 - n))


def byteswap32(x):
    return ((x & 0xff) << 24) | (((x >> 8) & 0xff) << 16) | \
           (((x >> 16) & 0xff) << 8) | ((x >> 24) & 0xff)


class TWI:
    def __init__(self):
        self.k_len = 0  # word32
        self.l_key = [0] * 40  # word32
        self.s_key = [0] * 4  # word32
        self.qt_gen = 0  # word32
        self.q_tab = [[0] * 256, [0] * 256]  # byte
        self.mt_gen = 0  # word32
        self.m_tab = [[0] * 256, [0] * 256, [0] * 256, [0] * 256]  # word32
        self.mk_tab = [[0] * 256, [0] * 256, [0] * 256, [0] * 256]  # word32


def byte(x, n):
    return (x >> (8 * n)) & 0xff


tab_5b = [0, 90, 180, 238]
tab_ef = [0, 238, 180, 90]
ror4 = [0, 8, 1, 9, 2, 10, 3, 11, 4, 12, 5, 13, 6, 14, 7, 15]
ashx = [0, 9, 2, 11, 4, 13, 6, 15, 8, 1, 10, 3, 12, 5, 14, 7]
qt0 = [[8, 1, 7, 13, 6, 15, 3, 2, 0, 11, 5, 9, 14, 12, 10, 4],
       [2, 8, 11, 13, 15, 7, 6, 14, 3, 1, 9, 4, 0, 10, 12, 5]]
qt1 = [[14, 12, 11, 8, 1, 2, 3, 5, 15, 4, 10, 6, 7, 0, 9, 13],
       [1, 14, 2, 11, 4, 12, 3, 7, 6, 13, 10, 5, 15, 9, 0, 8]]
qt2 = [[11, 10, 5, 14, 6, 13, 9, 0, 12, 8, 15, 3, 2, 4, 7, 1],
       [4, 12, 7, 5, 1, 6, 9, 10, 0, 14, 13, 8, 2, 11, 3, 15]]
qt3 = [[13, 7, 15, 4, 1, 2, 6, 14, 9, 11, 3, 0, 8, 5, 12, 10],
       [11, 9, 5, 1, 12, 3, 13, 14, 6, 4, 7, 15, 2, 0, 8, 10]]


def qp(n, x):  # word32, byte
    n %= 0x100000000
    x %= 0x100
    a0 = x >> 4
    b0 = x & 15
    a1 = a0 ^ b0
    b1 = ror4[b0] ^ ashx[a0]
    a2 = qt0[n][a1]
    b2 = qt1[n][b1]
    a3 = a2 ^ b2
    b3 = ror4[b2] ^ ashx[a2]
    a4 = qt2[n][a3]
    b4 = qt3[n][b3]
    return (b4 << 4) | a4


def gen_qtab(pkey):
    for i in range(256):
        pkey.q_tab[0][i] = qp(0, i)
        pkey.q_tab[1][i] = qp(1, i)


def gen_mtab(pkey):
    for i in range(256):
        f01 = pkey.q_tab[1][i]
        f01 = pkey.q_tab[1][i]
        f5b = ((f01) ^ ((f01) >> 2) ^ tab_5b[(f01) & 3])
        fef = ((f01) ^ ((f01) >> 1) ^ ((f01) >> 2) ^ tab_ef[(f01) & 3])
        pkey.m_tab[0][i] = f01 + (f5b << 8) + (fef << 16) + (fef << 24)
        pkey.m_tab[2][i] = f5b + (fef << 8) + (f01 << 16) + (fef << 24)

        f01 = pkey.q_tab[0][i]
        f5b = ((f01) ^ ((f01) >> 2) ^ tab_5b[(f01) & 3])
        fef = ((f01) ^ ((f01) >> 1) ^ ((f01) >> 2) ^ tab_ef[(f01) & 3])
        pkey.m_tab[1][i] = fef + (fef << 8) + (f5b << 16) + (f01 << 24)
        pkey.m_tab[3][i] = f5b + (f01 << 8) + (fef << 16) + (f5b << 24)


def gen_mk_tab(pkey, key):
    if pkey.k_len == 2:
        for i in range(256):
            by = i % 0x100
            pkey.mk_tab[0][i] = pkey.m_tab[0][pkey.q_tab[0][
                pkey.q_tab[0][by] ^ byte(key[1], 0)] ^
                byte(key[0], 0)]
            pkey.mk_tab[1][i] = pkey.m_tab[1][pkey.q_tab[0][
                pkey.q_tab[1][by] ^ byte(key[1], 1)] ^
                byte(key[0], 1)]
            pkey.mk_tab[2][i] = pkey.m_tab[2][pkey.q_tab[1][
                pkey.q_tab[0][by] ^ byte(key[1], 2)] ^
                byte(key[0], 2)]
            pkey.mk_tab[3][i] = pkey.m_tab[3][pkey.q_tab[1][
                pkey.q_tab[1][by] ^ byte(key[1], 3)] ^
                byte(key[0], 3)]
    if pkey.k_len == 3:
        for i in range(256):
            by = i % 0x100
            pkey.mk_tab[0][i] = pkey.m_tab[0][pkey.q_tab[0][
                pkey.q_tab[0][pkey.q_tab[1][by] ^
                              byte(key[2], 0)] ^ byte(key[1], 0)] ^
                byte(key[0], 0)]
            pkey.mk_tab[1][i] = pkey.m_tab[1][pkey.q_tab[0][
                pkey.q_tab[1][pkey.q_tab[1][by] ^
                              byte(key[2], 1)] ^ byte(key[1], 1)] ^
                byte(key[0], 1)]
            pkey.mk_tab[2][i] = pkey.m_tab[2][pkey.q_tab[1][
                pkey.q_tab[0][pkey.q_tab[0][by] ^
                              byte(key[2], 2)] ^ byte(key[1], 2)] ^
                byte(key[0], 2)]
            pkey.mk_tab[3][i] = pkey.m_tab[3][pkey.q_tab[1][
                pkey.q_tab[1][pkey.q_tab[0][by] ^
                              byte(key[2], 3)] ^ byte(key[1], 3)] ^
                byte(key[0], 3)]
    if pkey.k_len == 4:
        for i in range(256):
            by = i % 0x100
            pkey.mk_tab[0][i] = pkey.m_tab[0][pkey.q_tab[0][
                pkey.q_tab[0][pkey.q_tab[1][
                    pkey.q_tab[1][by] ^ byte(key[3], 0)] ^
                    byte(key[2], 0)] ^ byte(key[1], 0)] ^
                byte(key[0], 0)]
            pkey.mk_tab[1][i] = pkey.m_tab[1][pkey.q_tab[0][
                pkey.q_tab[1][pkey.q_tab[1][
                    pkey.q_tab[0][by] ^ byte(key[3], 1)] ^
                    byte(key[2], 1)] ^ byte(key[1], 1)] ^
                byte(key[0], 1)]
            pkey.mk_tab[2][i] = pkey.m_tab[2][pkey.q_tab[1][
                pkey.q_tab[0][pkey.q_tab[0][
                    pkey.q_tab[0][by] ^ byte(key[3], 2)] ^
                    byte(key[2], 2)] ^ byte(key[1], 2)] ^
                byte(key[0], 2)]
            pkey.mk_tab[3][i] = pkey.m_tab[3][pkey.q_tab[1][
                pkey.q_tab[1][pkey.q_tab[0][
                    pkey.q_tab[1][by] ^ byte(key[3], 3)] ^
                    byte(key[2], 3)] ^ byte(key[1], 3)] ^
                byte(key[0], 3)]


def h_fun(pkey, x, key):
    b0 = byte(x, 0)
    b1 = byte(x, 1)
    b2 = byte(x, 2)
    b3 = byte(x, 3)
    if pkey.k_len >= 4:
        b0 = pkey.q_tab[1][b0] ^ byte(key[3], 0)
        b1 = pkey.q_tab[0][b1] ^ byte(key[3], 1)
        b2 = pkey.q_tab[0][b2] ^ byte(key[3], 2)
        b3 = pkey.q_tab[1][b3] ^ byte(key[3], 3)
    if pkey.k_len >= 3:
        b0 = pkey.q_tab[1][b0] ^ byte(key[2], 0)
        b1 = pkey.q_tab[1][b1] ^ byte(key[2], 1)
        b2 = pkey.q_tab[0][b2] ^ byte(key[2], 2)
        b3 = pkey.q_tab[0][b3] ^ byte(key[2], 3)
    if pkey.k_len >= 2:
        b0 = pkey.q_tab[0][pkey.q_tab[0][b0] ^
                           byte(key[1], 0)] ^ byte(key[0], 0)
        b1 = pkey.q_tab[0][pkey.q_tab[1][b1] ^
                           byte(key[1], 1)] ^ byte(key[0], 1)
        b2 = pkey.q_tab[1][pkey.q_tab[0][b2] ^
                           byte(key[1], 2)] ^ byte(key[0], 2)
        b3 = pkey.q_tab[1][pkey.q_tab[1][b3] ^
                           byte(key[1], 3)] ^ byte(key[0], 3)
    return pkey.m_tab[0][b0] ^ pkey.m_tab[1][b1] ^ \
        pkey.m_tab[2][b2] ^ pkey.m_tab[3][b3]


def mds_rem(p0, p1):
    i, t, u = 0, 0, 0
    for i in range(8):
        t = p1 >> 24
        p1 = ((p1 << 8) & 0xffffffff) | (p0 >> 24)
        p0 = (p0 << 8) & 0xffffffff
        u = (t << 1) & 0xffffffff
        if t & 0x80:
            u ^= 0x0000014d
        p1 ^= t ^ ((u << 16) & 0xffffffff)
        u ^= (t >> 1)
        if t & 0x01:
            u ^= 0x0000014d >> 1
        p1 ^= ((u << 24) & 0xffffffff) | ((u << 8) & 0xffffffff)
    return p1


def set_key(pkey, in_key, key_len):
    pkey.qt_gen = 0
    if not pkey.qt_gen:
        gen_qtab(pkey)
        pkey.qt_gen = 1
    pkey.mt_gen = 0
    if not pkey.mt_gen:
        gen_mtab(pkey)
        pkey.mt_gen = 1
    pkey.k_len = int((key_len * 8) / 64)

    a = 0
    b = 0
    me_key = [0, 0, 0, 0]
    mo_key = [0, 0, 0, 0]
    for i in range(pkey.k_len):
        if WORD_BIGENDIAN:
            a = byteswap32(in_key[i + 1])
            me_key[i] = a
            b = byteswap32(in_key[i + i + 1])
        else:
            a = in_key[i + i]
            me_key[i] = a
            b = in_key[i + i + 1]
        mo_key[i] = b
        pkey.s_key[pkey.k_len - i - 1] = mds_rem(a, b)
    for i in range(0, 40, 2):
        a = (0x01010101 * i) % 0x100000000
        b = (a + 0x01010101) % 0x100000000
        a = h_fun(pkey, a, me_key)
        b = rotl32(h_fun(pkey, b, mo_key), 8)
        pkey.l_key[i] = (a + b) % 0x100000000
        pkey.l_key[i + 1] = rotl32((a + 2 * b) % 0x100000000, 9)
    gen_mk_tab(pkey, pkey.s_key)


def encrypt(pkey, in_blk):
    blk = [0, 0, 0, 0]

    if WORD_BIGENDIAN:
        blk[0] = byteswap32(in_blk[0]) ^ pkey.l_key[0]
        blk[1] = byteswap32(in_blk[1]) ^ pkey.l_key[1]
        blk[2] = byteswap32(in_blk[2]) ^ pkey.l_key[2]
        blk[3] = byteswap32(in_blk[3]) ^ pkey.l_key[3]
    else:
        blk[0] = in_blk[0] ^ pkey.l_key[0]
        blk[1] = in_blk[1] ^ pkey.l_key[1]
        blk[2] = in_blk[2] ^ pkey.l_key[2]
        blk[3] = in_blk[3] ^ pkey.l_key[3]

    for i in range(8):
        t1 = (pkey.mk_tab[0][byte(blk[1], 3)] ^
              pkey.mk_tab[1][byte(blk[1], 0)] ^
              pkey.mk_tab[2][byte(blk[1], 1)] ^
              pkey.mk_tab[3][byte(blk[1], 2)])
        t0 = (pkey.mk_tab[0][byte(blk[0], 0)] ^
              pkey.mk_tab[1][byte(blk[0], 1)] ^
              pkey.mk_tab[2][byte(blk[0], 2)] ^
              pkey.mk_tab[3][byte(blk[0], 3)])

        blk[2] = rotr32(blk[2] ^
                        ((t0 + t1 + pkey.l_key[4 * (i) + 8]) % 0x100000000), 1)
        blk[3] = rotl32(blk[3], 1) ^ \
            ((t0 + 2 * t1 + pkey.l_key[4 * (i) + 9]) % 0x100000000)

        t1 = (pkey.mk_tab[0][byte(blk[3], 3)] ^
              pkey.mk_tab[1][byte(blk[3], 0)] ^
              pkey.mk_tab[2][byte(blk[3], 1)] ^
              pkey.mk_tab[3][byte(blk[3], 2)])
        t0 = (pkey.mk_tab[0][byte(blk[2], 0)] ^
              pkey.mk_tab[1][byte(blk[2], 1)] ^
              pkey.mk_tab[2][byte(blk[2], 2)] ^
              pkey.mk_tab[3][byte(blk[2], 3)])

        blk[0] = rotr32(blk[0] ^
                        ((t0 + t1 + pkey.l_key[4 * (i) + 10]) %
                         0x100000000), 1)
        blk[1] = rotl32(blk[1], 1) ^ \
            ((t0 + 2 * t1 + pkey.l_key[4 * (i) + 11]) % 0x100000000)

    if WORD_BIGENDIAN:
        in_blk[0] = byteswap32(blk[2] ^ pkey.l_key[4])
        in_blk[1] = byteswap32(blk[3] ^ pkey.l_key[5])
        in_blk[2] = byteswap32(blk[0] ^ pkey.l_key[6])
        in_blk[3] = byteswap32(blk[1] ^ pkey.l_key[7])
    else:
        in_blk[0] = blk[2] ^ pkey.l_key[4]
        in_blk[1] = blk[3] ^ pkey.l_key[5]
        in_blk[2] = blk[0] ^ pkey.l_key[6]
        in_blk[3] = blk[1] ^ pkey.l_key[7]

    return


def decrypt(pkey, in_blk):
    blk = [0, 0, 0, 0]

    if WORD_BIGENDIAN:
        blk[0] = byteswap32(in_blk[0]) ^ pkey.l_key[4]
        blk[1] = byteswap32(in_blk[1]) ^ pkey.l_key[5]
        blk[2] = byteswap32(in_blk[2]) ^ pkey.l_key[6]
        blk[3] = byteswap32(in_blk[3]) ^ pkey.l_key[7]
    else:
        blk[0] = in_blk[0] ^ pkey.l_key[4]
        blk[1] = in_blk[1] ^ pkey.l_key[5]
        blk[2] = in_blk[2] ^ pkey.l_key[6]
        blk[3] = in_blk[3] ^ pkey.l_key[7]

    for i in range(7, -1, -1):
        t1 = (pkey.mk_tab[0][byte(blk[1], 3)] ^
              pkey.mk_tab[1][byte(blk[1], 0)] ^
              pkey.mk_tab[2][byte(blk[1], 1)] ^
              pkey.mk_tab[3][byte(blk[1], 2)])
        t0 = (pkey.mk_tab[0][byte(blk[0], 0)] ^
              pkey.mk_tab[1][byte(blk[0], 1)] ^
              pkey.mk_tab[2][byte(blk[0], 2)] ^
              pkey.mk_tab[3][byte(blk[0], 3)])

        blk[2] = rotl32(blk[2], 1) ^ ((t0 + t1 +
                                       pkey.l_key[4 * (i) + 10]) % 0x100000000)
        blk[3] = rotr32(blk[3] ^ ((t0 + 2 * t1 +
                                   pkey.l_key[4 * (i) + 11]) % 0x100000000), 1)

        t1 = (pkey.mk_tab[0][byte(blk[3], 3)] ^
              pkey.mk_tab[1][byte(blk[3], 0)] ^
              pkey.mk_tab[2][byte(blk[3], 1)] ^
              pkey.mk_tab[3][byte(blk[3], 2)])
        t0 = (pkey.mk_tab[0][byte(blk[2], 0)] ^
              pkey.mk_tab[1][byte(blk[2], 1)] ^
              pkey.mk_tab[2][byte(blk[2], 2)] ^
              pkey.mk_tab[3][byte(blk[2], 3)])

        blk[0] = rotl32(blk[0], 1) ^ ((t0 + t1 +
                                       pkey.l_key[4 * (i) + 8]) % 0x100000000)
        blk[1] = rotr32(blk[1] ^ ((t0 + 2 * t1 +
                                   pkey.l_key[4 * (i) + 9]) % 0x100000000), 1)

    if WORD_BIGENDIAN:
        in_blk[0] = byteswap32(blk[2] ^ pkey.l_key[0])
        in_blk[1] = byteswap32(blk[3] ^ pkey.l_key[1])
        in_blk[2] = byteswap32(blk[0] ^ pkey.l_key[2])
        in_blk[3] = byteswap32(blk[1] ^ pkey.l_key[3])
    else:
        in_blk[0] = blk[2] ^ pkey.l_key[0]
        in_blk[1] = blk[3] ^ pkey.l_key[1]
        in_blk[2] = blk[0] ^ pkey.l_key[2]
        in_blk[3] = blk[1] ^ pkey.l_key[3]
    return


__testkey = b'\xD4\x3B\xB7\x55\x6E\xA3\x2E\x46\xF2\xA2\x82\xB7\xD4' + \
            b'\x5B\x4E\x0D\x57\xFF\x73\x9D\x4D\xC9\x2C\x1B\xD7\xFC' + \
            b'\x01\x70\x0C\xC8\x21\x6F'

__testdat = b'\x90\xAF\xE9\x1B\xB2\x88\x54\x4F\x2C\x32\xDC\x23\x9B\x26\x35\xE6'

__testresult = b'l\xb4V\x1c@\xbf\n\x97\x05\x93\x1c\xb6\xd4\x08\xe7\xfa'

assert Twofish(__testkey).encrypt(__testdat) == __testresult

assert __testdat == Twofish(__testkey).decrypt(__testresult)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                "..")))

from pasaffe_lib.readdb import PassSafeFile  # noqa: E402


class TestPasswordGorilla15363(unittest.TestCase):
    def setUp(self):
        self.passfile = PassSafeFile(
            './tests/databases/pw-gorilla-15363.psafe3', 'pasaffe')

    def test_num_entries(self):
        self.assertEqual(len(self.passfile.records), 4)

    def test_empty_folders(self):
        # This version of Password Gorilla doesn't save empty folders
        self.assertEqual(len(self.passfile.empty_folders), 0)

    def test_get_database_version_string(self):
        self.assertEqual(self.passfile.get_database_version_string(), "03.00")

    def test_get_database_uuid(self):
        self.assertEqual(self.passfile.header[1],
                         b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                         b'\x00\x00\x00\x00\x00')

    def test_get_saved_name(self):
        self.assertEqual(self.passfile.get_saved_name(), None)

    def test_get_saved_host(self):
        self.assertEqual(self.passfile.get_saved_host(), None)

    def test_get_saved_application(self):
        self.assertEqual(self.passfile.get_saved_application(), None)

    def test_get_saved_date_string(self):
        self.assertEqual(self.passfile.get_saved_date_string(), None)

    def test_entry_1(self):
        uuid = '8642342dfd444e044f50ca8b3b12fd67'
        self.assertEqual(self.passfile.records[uuid][2], 'top\\.withdot')
        self.assertEqual(self.passfile.get_folder_list(uuid), ['top.withdot'])
        self.assertEqual(self.passfile.records[uuid][3], 'insidedot')
        self.assertEqual(self.passfile.records[uuid][4], 'usernamedot')
        self.assertFalse(5 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.records[uuid][6], 'passworddot')
        self.assertEqual(self.passfile.get_password_time(uuid, False),
                         'Thu, 25 Jul 2013 23:58:56')
        self.assertEqual(self.passfile.get_modification_time(uuid, False),
                         'Thu, 25 Jul 2013 23:58:56')

    def test_entry_2(self):
        uuid = '666b02ae0f574d4549f8f57ddae70d77'
        self.assertFalse(2 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.get_folder_list(uuid), [])
        self.assertEqual(self.passfile.records[uuid][3], 'toplevel1')
        self.assertEqual(self.passfile.records[uuid][4], 'usernametop')
        self.assertEqual(self.passfile.records[uuid][5], 'This is a note')
        self.assertEqual(self.passfile.records[uuid][6], 'passwordtop')
        self.assertEqual(self.passfile.get_password_time(uuid, False),
                         'Thu, 25 Jul 2013 23:58:06')
        self.assertEqual(self.passfile.get_modification_time(uuid, False),
                         'Thu, 25 Jul 2013 23:58:06')

    def test_entry_3(self):
        uuid = 'aff6f326353f4be86009bf342e0a0af7'
        self.assertEqual(self.passfile.records[uuid][2],
                         'topgroup1.topgroup2.topgroup3')
        self.assertEqual(self.passfile.get_folder_list(uuid),
                         ['topgroup1', 'topgroup2', 'topgroup3'])
        self.assertEqual(self.passfile.records[uuid][3], 'entrylevel3')
        self.assertEqual(self.passfile.records[uuid][4], 'username3')
        self.assertFalse(5 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.records[uuid][6], 'password3')
        self.assertEqual(self.passfile.get_password_time(uuid, False),
                         'Thu, 25 Jul 2013 00:33:33')
        self.assertEqual(self.passfile.get_modification_time(uuid, False),
                         'Thu, 25 Jul 2013 00:33:33')

    def test_entry_4(self):
        uuid = '66d1cf5e60dc424d6a3b19c89aed2f1e'
        self.assertEqual(self.passfile.records[uuid][2], 'topgroup1')
        self.assertEqual(self.passfile.get_folder_list(uuid), ['topgroup1'])
        self.assertEqual(self.passfile.records[uuid][3], 'entrylevel1')
        self.assertEqual(self.passfile.records[uuid][4], 'username1')
        self.assertEqual(self.passfile.records[uuid][5],
                         'This is a note\nThis is a second line\n'
                         'Unicode: éléphant')
        self.assertEqual(self.passfile.records[uuid][6], 'password1')
        self.assertEqual(self.passfile.get_password_time(uuid, False),
                         'Thu, 25 Jul 2013 00:33:06')
        self.assertEqual(self.passfile.get_modification_time(uuid, False),
                         'Thu, 25 Jul 2013 00:33:06')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
import tempfile
import shutil
import subprocess
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                                                 "..")))

from pasaffe_lib.readdb import PassSafeFile  # noqa: E402


class TestImportEntry(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.test_db = os.path.join(self.tempdir, 'test.psafe3')
        shutil.copy('./tests/databases/pasaffe-025.psafe3', self.test_db)

    def tearDown(self):
        if os.path.exists(self.tempdir):
            shutil.rmtree(self.tempdir)

    def test_orig_db(self):
        passfile = PassSafeFile(self.test_db, 'pasaffe')
        self.assertEqual(len(passfile.records), 3)

    def test_import_entry(self):

        name = 'testimport1'
        url = 'http://www.launchpad.net/pasaffe'
        user = 'testuser'
        password = 'testpass'
        note = "This is a note"

        rc = subprocess.call(['bin/pasaffe-import-entry', '-q',
                              '-f', self.test_db,
                              '-m', 'pasaffe',
                              '-e', name,
                              '-l', url,
                              '-u', user,
                              '-p', password,
                              '-n', note])

        self.assertEqual(rc, 0)

        passfile = PassSafeFile(self.test_db, 'pasaffe')
        self.assertEqual(len(passfile.records), 4)

        # Locate the new entry
        for uuid in passfile.records:
            if passfile.records[uuid][3] == name:
                break

        self.assertEqual(passfile.records[uuid][3], name)
        self.assertEqual(passfile.records[uuid][4], user)
        self.assertEqual(passfile.records[uuid][5], note)
        self.assertEqual(passfile.records[uuid][6], password)
        self.assertEqual(passfile.records[uuid][13], url)


if __name__ == '__main__':
    unittest.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from pasaffe_lib.AboutDialog import AboutDialog
from pasaffe_lib import get_version

logger = logging.getLogger('pasaffe')

# pylint: disable=E1101


# See pasaffe_lib.AboutDialog.py for more details about how this class works.
class AboutPasaffeDialog(AboutDialog):
    __gtype_name__ = "AboutPasaffeDialog"

    def finish_initializing(self, builder):  # pylint: disable=E1002
        """Set up the about dialog"""
        super(AboutPasaffeDialog, self).finish_initializing(builder)

        self.set_version(get_version())



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import struct
import hashlib
import os
import time
from binascii import hexlify

from . import blowfish
import logging
logger = logging.getLogger('pasaffe_lib')


class GPassFile:

    def __init__(self, filename=None, password=None):
        '''Reads a GPass file'''

        self.records = {}
        self.gpass_iv = b"\x05\x17\x01\x7b\x0c\x03\x36\x5e"
        self.decoded_db = None
        self.index = 0
        self.imported_folders = {}
        self.empty_folders = []

        self.cipher = None

        if filename is not None:
            self._readfile(filename, password)

    def _parse_entry(self, entry_id, parent_id, entry_data):
        '''Parse an entry'''

        uuid = os.urandom(16)
        uuid_hex = hexlify(uuid).decode('utf-8')
        timestamp = struct.pack("<I", int(time.time()))
        new_entry = {1: uuid, 3: '', 4: '', 6: '',
                     7: timestamp, 8: timestamp, 12: timestamp}

        # Stick the parent_id in here for now, we'll remove it
        # once we've figured out the folder tree
        new_entry[99] = parent_id

        entry_index = 0

        entry_index, entry_name = self.get_entry_string(entry_data,
                                                        entry_index)
        new_entry[3] = entry_name
        logger.debug("entry_name is %s" % entry_name)

        entry_index, entry_desc = self.get_entry_string(entry_data,
                                                        entry_index)
        if entry_desc != "":
            new_entry[5] = entry_desc
        logger.debug("entry_desc is %s" % entry_desc)

        entry_index, entry_ctime = self.get_entry_int(entry_data,
                                                      entry_index)
        new_entry[7] = entry_ctime
        logger.debug("creation time is %s" % hexlify(entry_ctime))

        entry_index, entry_mtime = self.get_entry_int(entry_data,
                                                      entry_index)

        # Gpass doesn't separately track when password was modified
        new_entry[8] = entry_mtime
        new_entry[12] = entry_mtime
        logger.debug("modification time is %s" %
                     hexlify(entry_mtime))

        entry_index, entry_expflag = self.get_entry_int(entry_data,
                                                        entry_index)
        logger.debug("expiration flag is %s" % hexlify(entry_expflag))

        entry_index, entry_etime = self.get_entry_int(entry_data,
                                                      entry_index)
        if struct.unpack("<I", entry_expflag)[0] != 0:
            new_entry[10] = entry_etime
        logger.debug("expiration time is %s" % hexlify(entry_etime))

        entry_index, entry_username = self.get_entry_string(entry_data,
                                                            entry_index)
        new_entry[4] = entry_username
        logger.debug("username is %s" % entry_username)

        entry_index, entry_pass = self.get_entry_string(entry_data,
                                                        entry_index)
        new_entry[6] = entry_pass
        # logger.debug("password is %s" % entry_pass)

        entry_index, entry_url = self.get_entry_string(entry_data,
                                                       entry_index)
        if entry_url != "":
            new_entry[13] = entry_url
        logger.debug("URL is %s" % entry_url)

        self.records[uuid_hex] = new_entry

    def _parse_folder(self, entry_id, parent_id, entry_data):
        '''Parse a folder'''

        entry_index = 0
        entry_index, entry_name = self.get_entry_string(entry_data,
                                                        entry_index)
        logger.debug("folder_name is %s" % entry_name)

        folder = [entry_name, parent_id]
        self.imported_folders[entry_id] = folder

        # We don't actually do anything with these, but leave them
        # here in case we do parse them at some point
        #
        # entry_index, entry_desc = self.get_entry_string(entry_data,
        #                                                entry_index)
        # logger.debug("entry_desc is %s" % entry_desc)
        #
        # entry_index, entry_ctime = self.get_entry_int(entry_data,
        #                                              entry_index)
        # logger.debug("creation time is %s" % hexlify(entry_ctime))
        #
        # entry_index, entry_mtime = self.get_entry_int(entry_data,
        #                                              entry_index)
        # logger.debug("modification time is %s" %
        #             hexlify(entry_mtime))
        #
        # entry_index, entry_expflag = self.get_entry_int(entry_data,
        #                                                entry_index)
        # logger.debug("expiration flag is %s" % hexlify(entry_expflag))
        #
        # entry_index, entry_etime = self.get_entry_int(entry_data,
        #                                                  entry_index)
        # logger.debug("expiration time is %s" % hexlify(entry_etime))

    def _readfile(self, filename, password):
        '''Parses database file'''
        try:
            dbfile = open(filename, 'rb')
        except Exception:
            raise RuntimeError("Could not open %s. Aborting." % filename)

        dbdata = dbfile.read()

        key = hashlib.sha1(password.encode('utf-8')).digest()
        self.cipher = blowfish.Blowfish(key)
        self.cipher.initCBC(iv=struct.unpack("Q", self.gpass_iv)[0])

        self.decoded_db = self.cipher.decryptCBC(dbdata)
        self.remove_padding()
        self.validate_header()

        self.index = 0

        while self.index < len(self.decoded_db):

            # We just generate a new UUID instead of using the GPass id
            entry_id = self.get_int()
            logger.debug("id is %s" % hexlify(entry_id))

            # We don't handle parents in Pasaffe for now
            parent_id = self.get_int()
            logger.debug("parent_id is %s" % hexlify(parent_id))

            entry_type = self.get_bytes()
            logger.debug("entry_type is %s" % entry_type)

            entry_data = self.get_bytes()

            # We don't handle anything else than standard entries for now
            if entry_type == b"general":
                self._parse_entry(entry_id, parent_id, entry_data)
            elif entry_type == b"folder":
                self._parse_folder(entry_id, parent_id, entry_data)
            else:
                continue

        dbfile.close()

        self._set_folders()
        self._find_empty_folders()

    def _folder_list_to_field(self, folder_list):
        '''Converts a folder list to a folder field'''
        field = ""

        if folder_list is None:
            return field

        for folder in folder_list:
            if field != "":
                field += "."
            field += folder.replace(".", "\\.")
        return field

    def _set_folders(self):
        '''Set an entry's folder from the folder list'''

        for uuid in self.records:
            if self.records[uuid][99] != b'\x00\x00\x00\x00':
                folder_list = self._get_folder_list(self.records[uuid][99])
                self.records[uuid][2] = self._folder_list_to_field(folder_list)

            del self.records[uuid][99]

    def _find_empty_folders(self):
        '''Find empty folders'''

        for folder in self.imported_folders:
            name, parent = self.imported_folders[folder]
            folder_list = [name]
            while(parent != b'\x00\x00\x00\x00'):
                name, parent = self.imported_folders[parent]
                folder_list.insert(0, name)

            self.add_empty_folder(folder_list)

    def add_empty_folder(self, folder):
        '''Adds a folder to the empty folders list'''

        if folder is None or folder == []:
            return

        for part in range(len(folder)):
            field = self._folder_list_to_field(folder[:part + 1])
            logger.debug("searching for %s" % field)
            if field not in self.empty_folders:
                # Make sure it's actually empty
                found = False
                for uuid in list(self.records.keys()):
                    if 2 not in self.records[uuid]:
                        continue
                    if self.records[uuid][2] == field:
                        logger.debug("folder %s isn't empty" % field)
                        found = True
                        break

                if found is False:
                    logger.debug("adding %s" % field)
                    self.empty_folders.append(field)

    def _get_folder_list(self, parent_id):
        '''From a parent id, figure out a folder list'''

        name, parent = self.imported_folders[parent_id]
        folder_list = [name]
        while(parent != b'\x00\x00\x00\x00'):
            name, parent = self.imported_folders[parent]
            folder_list.insert(0, name)

        return folder_list

    def remove_padding(self):
        padding = self.decoded_db[-1]

        for byte in self.decoded_db[-padding:]:
            if byte != padding:
                return

        self.decoded_db = self.decoded_db[:-padding]

    def validate_header(self):
        gpass_magic_prefix = b"GPassFile version 1.1.0"

        logger.debug("decoded_db header is %s" %
                     self.decoded_db[:len(gpass_magic_prefix)])

        # Check magic prefix
        if not self.decoded_db.startswith(gpass_magic_prefix):
            raise RuntimeError("File is not a GPass database,"
                               " or wrong password. Aborting.")

        # Skip header
        self.decoded_db = self.decoded_db[len(gpass_magic_prefix):]

    def get_int(self):
        value = self.decoded_db[self.index:self.index + 4]
        self.index += 4
        return value

    def get_bytes(self):
        length = struct.unpack("<I", self.get_int())[0]
        value = self.decoded_db[self.index:self.index + length]
        self.index += length
        return value

    def get_entry_int(self, entry, entry_index):
        value = 0
        b = 1

        for i in range(6):
            c = entry[entry_index + i]
            if c & 0x80:
                value += b * (c & 0x7f)
                b *= 0x80
            else:
                value += b * c
                return entry_index + i + 1, struct.pack("<I", value)

    def get_entry_string(self, entry, entry_index):
        entry_index, length = self.get_entry_int(entry, entry_index)
        length = struct.unpack("<I", length)[0]
        value = entry[entry_index:entry_index + length]
        entry_index += length
        return entry_index, value.decode('utf-8')



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2014 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                "..")))

from pasaffe_lib.readdb import PassSafeFile  # noqa: E402


class TestPasswdSafeAndroid510(unittest.TestCase):
    def setUp(self):
        self.passfile = PassSafeFile(
            './tests/databases/passwdsafe-510.psafe3', 'pasaffe')

    def test_num_entries(self):
        self.assertEqual(len(self.passfile.records), 4)

    def test_empty_folders(self):
        self.assertEqual(len(self.passfile.empty_folders), 0)

    def test_get_database_version_string(self):
        self.assertEqual(self.passfile.get_database_version_string(), "03.0d")

    def test_get_database_uuid(self):
        self.assertEqual(self.passfile.header[1],
                         b'b\xfb\x00\x00\x9al2\x13\x05\xb8\xccnh\x81L\xb5')

    def test_get_saved_name(self):
        self.assertEqual(self.passfile.get_saved_name(),
                         'User')

    def test_get_saved_host(self):
        self.assertEqual(self.passfile.get_saved_host(),
                         'Galaxy Nexus')

    def test_get_saved_application(self):
        self.assertEqual(self.passfile.get_saved_application(),
                         'PasswdSafe 5.1.0')

    def test_get_saved_date_string(self):
        self.assertEqual(self.passfile.get_saved_date_string(),
                         'Sat, 25 Oct 2014 12:05:42')

    def test_entry_1(self):
        uuid = 'af5304edb06c32131d502487eeb21009'
        self.assertEqual(self.passfile.records[uuid][2], 'Toplevel2')
        self.assertEqual(self.passfile.get_folder_list(uuid), ['Toplevel2'])
        self.assertEqual(self.passfile.records[uuid][3], 'Intoplevel2')
        self.assertEqual(self.passfile.records[uuid][4], 'username')
        self.assertFalse(5 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.records[uuid][6], 'password')
        self.assertFalse(8 in self.passfile.records[uuid])
        self.assertFalse(12 in self.passfile.records[uuid])

    def test_entry_2(self):
        uuid = '7f1f06b3a36c3213061dc476b88f83fd'
        self.assertFalse(2 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.get_folder_list(uuid), [])
        self.assertEqual(self.passfile.records[uuid][3], 'Toplevel')
        self.assertEqual(self.passfile.records[uuid][4], 'topuser')
        self.assertEqual(self.passfile.records[uuid][5], 'This is a note\n' +
                                                         'This is line 2\n' +
                                                         'Unicode: éléphant')
        self.assertEqual(self.passfile.records[uuid][6], 'toppass')
        self.assertFalse(8 in self.passfile.records[uuid])
        self.assertFalse(12 in self.passfile.records[uuid])

    def test_entry_3(self):
        uuid = '82bd036bb86c32130facf0a8e3e3809b'
        self.assertEqual(self.passfile.records[uuid][2], 'Top.withdot')
        self.assertEqual(self.passfile.get_folder_list(uuid),
                         ['Top', 'withdot'])
        self.assertEqual(self.passfile.records[uuid][3], 'Indot')
        self.assertEqual(self.passfile.records[uuid][4], '')
        self.assertFalse(5 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.records[uuid][6], 'password')
        self.assertFalse(8 in self.passfile.records[uuid])
        self.assertFalse(12 in self.passfile.records[uuid])

    def test_entry_4(self):
        uuid = '116301c0aa6c321334769a2131e417c5'
        self.assertEqual(self.passfile.records[uuid][2], 'Toplevel1')
        self.assertEqual(self.passfile.get_folder_list(uuid), ['Toplevel1'])
        self.assertEqual(self.passfile.records[uuid][3], 'Intoplevel1')
        self.assertEqual(self.passfile.records[uuid][4], 'username')
        self.assertEqual(self.passfile.records[uuid][6], 'password')
        self.assertEqual(self.passfile.records[uuid][13],
                         'http://www.example.com')
        self.assertFalse(8 in self.passfile.records[uuid])
        self.assertFalse(12 in self.passfile.records[uuid])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# DO NOT TOUCH THIS (HEAD TO THE SECOND PART)

import os
import sys
from glob import glob

try:
    import DistUtilsExtra.auto
except ImportError:
    print('To build pasaffe you need ' +
          'https://launchpad.net/python-distutils-extra', file=sys.stderr)
    sys.exit(1)
assert DistUtilsExtra.auto.__version__ >= '2.18', \
    'needs DistUtilsExtra.auto >= 2.18'


def update_config(values={}):

    oldvalues = {}
    try:
        fin = open('pasaffe_lib/pasaffeconfig.py', 'r')
        fout = open(fin.name + '.new', 'w')

        for line in fin:
            fields = line.split(' = ')  # Separate variable from value
            if fields[0] in values:
                oldvalues[fields[0]] = fields[1].strip()
                line = "%s = %s\n" % (fields[0], values[fields[0]])
            fout.write(line)

        fout.flush()
        fout.close()
        fin.close()
        os.rename(fout.name, fin.name)
    except (OSError, IOError):
        print("ERROR: Can't find pasaffe_lib/pasaffeconfig.py")
        sys.exit(1)
    return oldvalues


class InstallAndUpdateDataDirectory(DistUtilsExtra.auto.install_auto):
    def run(self):
        values = {'__pasaffe_data_directory__': "'%s'" %
                  (self.prefix + '/share/pasaffe/'),
                  '__version__': "'%s'" % self.distribution.get_version()}

        # Older DistUtilsExtra put help files in /usr/share/gnome/help and
        # needed a ghelp: URL
        if DistUtilsExtra.auto.__version__ < '2.38':
            values['__help_prefix__'] = "'ghelp:'"
            values['__help_separator__'] = "'#'"

        previous_values = update_config(values)
        DistUtilsExtra.auto.install_auto.run(self)
        update_config(previous_values)


#
# YOU SHOULD MODIFY ONLY WHAT IS BELOW
#

DistUtilsExtra.auto.setup(
    name='pasaffe',
    version='0.56',
    license='GPL-3',
    author='Marc Deslauriers',
    author_email='marc.deslauriers@canonical.com',
    description='Password manager for GNOME',
    long_description='Pasaffe is an easy to use password manager for GNOME.',
    url='https://launchpad.net/pasaffe',
    data_files=[('share/mime/packages', glob('mime/*'))],
    cmdclass={'install': InstallAndUpdateDataDirectory}
    )



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
import subprocess


class TestPylint(unittest.TestCase):
    def test_project_errors_only(self):
        '''run pylint in error only mode

        your code may well work even with pylint errors
        but have some unusual code'''
        subprocess.call(["pylint", '-E', 'pasaffe'])


if __name__ == '__main__':
    'you will get better results with nosetests'
    unittest.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk  # noqa: E402

from . helpersgui import get_builder  # noqa: E402


class AboutDialog(Gtk.AboutDialog):
    __gtype_name__ = "AboutDialog"

    def __new__(cls):
        """Special static method that's automatically called by Python when
        constructing a new instance of this class.

        Returns a fully instantiated AboutDialog object.
        """
        builder = get_builder('AboutPasaffeDialog')
        new_object = builder.get_object("about_pasaffe_dialog")
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called while initializing this instance in __new__

        finish_initalizing should be called after parsing the ui definition
        and creating a AboutDialog object with it in order
        to finish initializing the start of the new AboutPasaffeDialog
        instance.

        Put your initialization code in here and leave __init__ undefined.
        """
        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
import time
import struct
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                                                 "..")))

from pasaffe_lib.figaroxml import FigaroXML  # noqa: E402


class TestFigaroXML079(unittest.TestCase):
    def setUp(self):
        self.passfile = FigaroXML('./tests/databases/figaro-079.xml')

    def _find_uuid(self, name):
        for uuid in self.passfile.records:
            if self.passfile.records[uuid][3] == name:
                return uuid
        return None

    def _get_time(self, uuid, entry):
        '''Returns a string of time'''
        if entry in self.passfile.records[uuid]:
            unpacked_time = struct.unpack(
                "<I", self.passfile.records[uuid][entry])[0]
            converted_time = time.gmtime(unpacked_time)
            return time.strftime("%a, %d %b %Y %H:%M:%S", converted_time)
        else:
            return None

    def test_num_entries(self):
        self.assertEqual(len(self.passfile.records), 3)

    def test_entry_1(self):

        entry_name = "topentry1"
        uuid = self._find_uuid(entry_name)

        self.assertFalse(2 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.records[uuid][3], entry_name)
        self.assertEqual(self.passfile.records[uuid][4], 'username1')
        self.assertEqual(self.passfile.records[uuid][5],
                         'This is a note\nSecond line\nUnicode: ééé\n')
        self.assertEqual(self.passfile.records[uuid][6], 'password1')
        self.assertEqual(self.passfile.records[uuid][13], 'hostname1')

    def test_entry_2(self):

        entry_name = "topentry2"
        uuid = self._find_uuid(entry_name)

        self.assertFalse(2 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.records[uuid][3], entry_name)
        self.assertEqual(self.passfile.records[uuid][4], 'username2')
        self.assertEqual(self.passfile.records[uuid][5], 'This is a note')
        self.assertEqual(self.passfile.records[uuid][6], 'password2')
        self.assertEqual(self.passfile.records[uuid][13], 'hostname2')

    def test_entry_3(self):

        entry_name = "topcategory3"
        uuid = self._find_uuid(entry_name)

        self.assertEqual(self.passfile.records[uuid][2], 'category1')
        self.assertEqual(self.passfile.records[uuid][3], entry_name)
        self.assertEqual(self.passfile.records[uuid][4], 'username3')
        self.assertEqual(self.passfile.records[uuid][5],
                         'This is notes\nLine 2')
        self.assertEqual(self.passfile.records[uuid][6], 'password3')
        self.assertEqual(self.passfile.records[uuid][13], 'hostname3')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
import subprocess
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                                                 "..")))


class TestDumpDB(unittest.TestCase):
    def setUp(self):
        pass

    def test_dump_db(self):
        out = subprocess.check_output(['bin/pasaffe-dump-db', '-q', '-f',
                                       './tests/databases/pasaffe-025.psafe3',
                                       '-m', 'pasaffe'])

        expected_out = b'''Entry: level1entry
Username: username1
Password: password1
Notes: This is a note

Entry: level3entry
Username: usernamelevel3
Password: passwordlevel3

Entry: topentry1
Username: username1
Password: password1
URL: http://www.example.com
Notes: This is a note

'''
        self.assertEqual(out, expected_out)

    def test_dump_db_verbose(self):
        out = subprocess.check_output(['bin/pasaffe-dump-db', '-f',
                                       './tests/databases/pasaffe-025.psafe3',
                                       '-m', 'pasaffe'])

        self.assertTrue(b'WARNING: this will display all password entries.'
                        in out)


if __name__ == '__main__':
    unittest.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gio
import logging
from pasaffe_lib.PreferencesDialog import PreferencesDialog

logger = logging.getLogger('pasaffe')


class PreferencesPasaffeDialog(PreferencesDialog):
    __gtype_name__ = "PreferencesPasaffeDialog"

    def finish_initializing(self, builder):  # pylint: disable=E1002
        """Set up the preferences dialog"""
        super(PreferencesPasaffeDialog, self).finish_initializing(builder)

        # Bind each preference widget to gsettings
        settings = Gio.Settings.new("net.launchpad.pasaffe")
        widget = self.builder.get_object('visible-secrets')
        settings.bind("visible-secrets", widget, "active",
                      Gio.SettingsBindFlags.DEFAULT)
        widget = self.builder.get_object('only-passwords-are-secret')
        settings.bind("only-passwords-are-secret", widget, "active",
                      Gio.SettingsBindFlags.DEFAULT)
        widget = self.builder.get_object('lock-on-idle')
        settings.bind("lock-on-idle", widget, "active",
                      Gio.SettingsBindFlags.DEFAULT)
        widget = self.builder.get_object('idle-timeout')
        settings.bind("idle-timeout", widget, "value",
                      Gio.SettingsBindFlags.DEFAULT)
        widget = self.builder.get_object('auto-save')
        settings.bind("auto-save", widget, "active",
                      Gio.SettingsBindFlags.DEFAULT)
        widget = self.builder.get_object('password-length')
        settings.bind("password-length", widget, "value",
                      Gio.SettingsBindFlags.DEFAULT)
        widget = self.builder.get_object('tree-expansion')
        settings.bind("tree-expansion", widget, "active-id",
                      Gio.SettingsBindFlags.DEFAULT)
        widget = self.builder.get_object('double-click')
        settings.bind("double-click", widget, "active-id",
                      Gio.SettingsBindFlags.DEFAULT)
        widget = self.builder.get_object('display-usernames')
        settings.bind("display-usernames", widget, "active",
                      Gio.SettingsBindFlags.DEFAULT)

        # Code for other initialization actions should be added here.
